@@ -2,12 +2,10 @@
 
 #include "assert.h"
 #include "byte.h"
-#include "tracing.h"
 
 /* Current encoding format version. */
 #define ENCODING_FORMAT 1
 
-
 void configurationInit(struct raft_configuration *c)
 {
     c->servers = NULL;
@@ -63,6 +61,25 @@ unsigned configurationIndexOfVoter(const struct raft_configuration *c,
     return c->n;
 }
 
+unsigned configurationActualIndexOfVoter(const struct raft_configuration *c,
+                                         unsigned i)
+{
+    unsigned j;
+    unsigned k = 0;
+    assert(c != NULL);
+
+    for (j = 0; j < c->n; j++) {
+        if (c->servers[j].role == RAFT_VOTER) {
+            if (k == i) {
+                return j;
+            }
+            k++;
+        }
+    }
+
+    return c->n;
+}
+
 const struct raft_server *configurationGet(const struct raft_configuration *c,
                                            const raft_id id)
 {
@@ -100,63 +117,87 @@ int configurationCopy(const struct raft_configuration *src,
 {
     size_t i;
     int rv;
+
     configurationInit(dst);
     for (i = 0; i < src->n; i++) {
         struct raft_server *server = &src->servers[i];
         rv = configurationAdd(dst, server->id, server->address, server->role);
         if (rv != 0) {
-            return rv;
+            goto err;
         }
     }
+
     return 0;
+
+err:
+    configurationClose(dst);
+    assert(rv == RAFT_NOMEM);
+    return rv;
 }
 
 int configurationAdd(struct raft_configuration *c,
-                     raft_id id,
+                     const raft_id id,
                      const char *address,
-                     int role)
+                     const int role)
 {
     struct raft_server *servers;
     struct raft_server *server;
+    char *address_copy;
     size_t i;
+    int rv;
     assert(c != NULL);
     assert(id != 0);
 
     if (role != RAFT_STANDBY && role != RAFT_VOTER && role != RAFT_SPARE) {
-        return RAFT_BADROLE;
+        rv = RAFT_BADROLE;
+        goto err;
     }
 
     /* Check that neither the given id or address is already in use */
     for (i = 0; i < c->n; i++) {
         server = &c->servers[i];
         if (server->id == id) {
-            return RAFT_DUPLICATEID;
+            rv = RAFT_DUPLICATEID;
+            goto err;
         }
         if (strcmp(server->address, address) == 0) {
-            return RAFT_DUPLICATEADDRESS;
+            rv = RAFT_DUPLICATEADDRESS;
+            goto err;
         }
     }
 
+    /* Make a copy of the given address */
+    address_copy = raft_malloc(strlen(address) + 1);
+    if (address_copy == NULL) {
+        rv = RAFT_NOMEM;
+        goto err;
+    }
+    strcpy(address_copy, address);
+
     /* Grow the servers array.. */
     servers = raft_realloc(c->servers, (c->n + 1) * sizeof *server);
     if (servers == NULL) {
-        return RAFT_NOMEM;
+        rv = RAFT_NOMEM;
+        goto err_after_address_copy;
     }
     c->servers = servers;
 
     /* Fill the newly allocated slot (the last one) with the given details. */
     server = &servers[c->n];
     server->id = id;
-    server->address = raft_malloc(strlen(address) + 1);
-    if (server->address == NULL) {
-        return RAFT_NOMEM;
-    }
-    strcpy(server->address, address);
+    server->address = address_copy;
     server->role = role;
 
     c->n++;
 
     return 0;
+
+err_after_address_copy:
+    raft_free(address_copy);
+err:
+    assert(rv == RAFT_BADROLE || rv == RAFT_DUPLICATEID ||
+           rv == RAFT_DUPLICATEADDRESS || rv == RAFT_NOMEM);
+    return rv;
 }
 
 int configurationRemove(struct raft_configuration *c, const raft_id id)
@@ -164,28 +205,30 @@ int configurationRemove(struct raft_configuration *c, const raft_id id)
     unsigned i;
     unsigned j;
     struct raft_server *servers;
+    int rv;
+
     assert(c != NULL);
 
     i = configurationIndexOf(c, id);
     if (i == c->n) {
-        return RAFT_BADID;
+        rv = RAFT_BADID;
+        goto err;
     }
 
     assert(i < c->n);
 
     /* If this is the last server in the configuration, reset everything. */
     if (c->n - 1 == 0) {
-        raft_free(c->servers[0].address);
-        raft_free(c->servers);
-        c->n = 0;
-        c->servers = NULL;
-        return 0;
+        assert(i == 0);
+        servers = NULL;
+        goto out;
     }
 
     /* Create a new servers array. */
     servers = raft_calloc(c->n - 1, sizeof *servers);
     if (servers == NULL) {
-        return RAFT_NOMEM;
+        rv = RAFT_NOMEM;
+        goto err;
     }
 
     /* Copy the first part of the servers array into a new array, excluding the
@@ -199,6 +242,7 @@ int configurationRemove(struct raft_configuration *c, const raft_id id)
         servers[j - 1] = c->servers[j];
     }
 
+out:
     /* Release the address of the server that was deleted. */
     raft_free(c->servers[i].address);
 
@@ -209,6 +253,10 @@ int configurationRemove(struct raft_configuration *c, const raft_id id)
     c->n--;
 
     return 0;
+
+err:
+    assert(rv == RAFT_BADID || rv == RAFT_NOMEM);
+    return rv;
 }
 
 size_t configurationEncodedSize(const struct raft_configuration *c)
@@ -236,19 +284,19 @@ size_t configurationEncodedSize(const struct raft_configuration *c)
 
 void configurationEncodeToBuf(const struct raft_configuration *c, void *buf)
 {
-    void *cursor = buf;
+    uint8_t *cursor = buf;
     unsigned i;
 
     /* Encoding format version */
     bytePut8(&cursor, ENCODING_FORMAT);
 
     /* Number of servers. */
-    bytePut64Unaligned(&cursor, c->n); /* cursor might not be 8-byte aligned */
+    bytePut64(&cursor, c->n);
 
     for (i = 0; i < c->n; i++) {
         struct raft_server *server = &c->servers[i];
         assert(server->address != NULL);
-        bytePut64Unaligned(&cursor, server->id); /* might not be aligned */
+        bytePut64(&cursor, server->id);
         bytePutString(&cursor, server->address);
         assert(server->role < 255);
         bytePut8(&cursor, (uint8_t)server->role);
@@ -258,6 +306,8 @@ void configurationEncodeToBuf(const struct raft_configuration *c, void *buf)
 int configurationEncode(const struct raft_configuration *c,
                         struct raft_buffer *buf)
 {
+    int rv;
+
     assert(c != NULL);
     assert(buf != NULL);
 
@@ -267,20 +317,26 @@ int configurationEncode(const struct raft_configuration *c,
     buf->len = configurationEncodedSize(c);
     buf->base = raft_malloc(buf->len);
     if (buf->base == NULL) {
-        return RAFT_NOMEM;
+        rv = RAFT_NOMEM;
+        goto err;
     }
 
     configurationEncodeToBuf(c, buf->base);
 
     return 0;
+
+err:
+    assert(rv == RAFT_NOMEM);
+    return rv;
 }
 
 int configurationDecode(const struct raft_buffer *buf,
                         struct raft_configuration *c)
 {
-    const void *cursor;
+    const uint8_t *cursor;
     size_t i;
     size_t n;
+    int rv;
 
     assert(c != NULL);
     assert(buf != NULL);
@@ -288,36 +344,35 @@ int configurationDecode(const struct raft_buffer *buf,
     /* TODO: use 'if' instead of assert for checking buffer boundaries */
     assert(buf->len > 0);
 
-    /* Check that the target configuration is empty. */
-    assert(c->n == 0);
-    assert(c->servers == NULL);
+    configurationInit(c);
 
     cursor = buf->base;
 
     /* Check the encoding format version */
     if (byteGet8(&cursor) != ENCODING_FORMAT) {
-        return RAFT_MALFORMED;
+        rv = RAFT_MALFORMED;
+        goto err;
     }
 
     /* Read the number of servers. */
-    n = (size_t)byteGet64Unaligned(&cursor);
+    n = (size_t)byteGet64(&cursor);
 
     /* Decode the individual servers. */
     for (i = 0; i < n; i++) {
         raft_id id;
         const char *address;
         int role;
-        int rv;
 
         /* Server ID. */
-        id = byteGet64Unaligned(&cursor);
+        id = byteGet64(&cursor);
 
         /* Server Address. */
         address = byteGetString(
             &cursor,
             buf->len - (size_t)((uint8_t *)cursor - (uint8_t *)buf->base));
         if (address == NULL) {
-            return RAFT_MALFORMED;
+            rv = RAFT_MALFORMED;
+            goto err;
         }
 
         /* Role code. */
@@ -325,28 +380,20 @@ int configurationDecode(const struct raft_buffer *buf,
 
         rv = configurationAdd(c, id, address, role);
         if (rv != 0) {
-            return rv;
+            /* Only valid configurations should be ever be encoded, so in case
+             * configurationAdd() fails because of invalid data we return
+             * RAFT_MALFORMED. */
+            if (rv != RAFT_NOMEM) {
+                rv = RAFT_MALFORMED;
+            }
+            goto err;
         }
     }
 
     return 0;
-}
 
-#define tracef(...) Tracef(r->tracer, __VA_ARGS__)
-void configurationTrace(const struct raft *r, struct raft_configuration *c, const char *msg)
-{
-    if (r == NULL || c == NULL || !r->tracer->enabled ) {
-        return;
-    }
-
-    tracef("%s", msg);
-    tracef("=== CONFIG START ===");
-    unsigned i;
-    struct raft_server *s;
-    for (i = 0; i < c->n; i++) {
-        s = &c->servers[i];
-        tracef("id:%llu address:%s role:%d", s->id, s->address, s->role);
-    }
-    tracef("=== CONFIG END ===");
+err:
+    assert(rv == RAFT_MALFORMED || rv == RAFT_NOMEM);
+    configurationClose(c);
+    return rv;
 }
-#undef tracef
