@@ -7,9 +7,12 @@
 
 #include "assert.h"
 #include "configuration.h"
+#include "convert.h"
 #include "entry.h"
+#include "legacy.h"
 #include "log.h"
 #include "queue.h"
+#include "random.h"
 #include "snapshot.h"
 #include "tracing.h"
 
@@ -30,6 +33,35 @@
  * instance. This should be enough for testing purposes. */
 #define MAX_PEERS 8
 
+struct raft_fixture_server
+{
+    bool alive;                /* If false, the server is down. */
+    raft_id id;                /* Server ID. */
+    char address[16];          /* Server address (stringified ID). */
+    struct raft_tracer tracer; /* Tracer. */
+    struct raft_io io;         /* In-memory raft_io implementation. */
+    struct raft raft;          /* Raft instance. */
+};
+
+struct raft_fixture_event
+{
+    unsigned server_index; /* Index of the server the event occurred on. */
+    int type;              /* Type of the event. */
+};
+
+RAFT_API int raft_fixture_event_type(struct raft_fixture_event *event)
+{
+    assert(event != NULL);
+    return event->type;
+}
+
+RAFT_API unsigned raft_fixture_event_server_index(
+    struct raft_fixture_event *event)
+{
+    assert(event != NULL);
+    return event->server_index;
+}
+
 /* Fields common across all request types. */
 #define REQUEST                                                            \
     int type;                  /* Request code type. */                    \
@@ -73,13 +105,6 @@ struct snapshot_put
     const struct raft_snapshot *snapshot;
 };
 
-/* Pending request to perform general work. */
-struct async_work
-{
-    REQUEST;
-    struct raft_io_async_work *req;
-};
-
 /* Pending request to load a snapshot. */
 struct snapshot_get
 {
@@ -120,6 +145,7 @@ struct io
     struct raft_snapshot *snapshot; /* Latest snapshot */
     struct raft_entry *entries;     /* Array or persisted entries */
     size_t n;                       /* Size of the persisted entries array */
+    raft_index start_index;
 
     /* Parameters passed via raft_io->init and raft_io->start */
     raft_id id;
@@ -136,10 +162,25 @@ struct io
     struct peer peers[MAX_PEERS];
     unsigned n_peers;
 
-    unsigned randomized_election_timeout; /* Value returned by io->random() */
-    unsigned network_latency;             /* Milliseconds to deliver RPCs */
-    unsigned disk_latency;                /* Milliseconds to perform disk I/O */
-    unsigned work_duration;               /* Milliseconds to long running work */
+    /* The randomized_election_timeout field stores the value that the raft
+     * instance will obtain the next time it calls RandomWithinRange() to obtain
+     * a random number in the [election_timeout, election_timeout * 2] range. We
+     * do that by passing raft_seed() a value that makes the pseudor random
+     * number generator produce exactly randomized_election_timeout. That value
+     * is what we store in the seed field below. Since calculating the seed that
+     * matches the desired randomized_election_timeout is somehow expensive, we
+     * also use randomized_election_timeout_prev to store the previous value of
+     * randomized_election_timeout, in order to re-use the same seed if nothing
+     * has changed.
+     *
+     * See serverSeed for more details. */
+    unsigned randomized_election_timeout;
+    unsigned randomized_election_timeout_prev;
+    unsigned seed;
+
+    unsigned network_latency; /* Milliseconds to deliver RPCs */
+    unsigned disk_latency;    /* Milliseconds to perform disk I/O */
+    unsigned work_duration;   /* Milliseconds to run async work */
 
     struct
     {
@@ -224,6 +265,13 @@ static void ioFlushAppend(struct io *s, struct append *append)
 {
     struct raft_entry *entries;
     unsigned i;
+    int status = 0;
+
+    /* Simulates a disk write failure. */
+    if (ioFaultTick(s)) {
+        status = RAFT_IOERR;
+        goto done;
+    }
 
     /* Allocate an array for the old entries plus the new ones. */
     entries = raft_realloc(s->entries, (s->n + append->n) * sizeof *s->entries);
@@ -240,8 +288,9 @@ static void ioFlushAppend(struct io *s, struct append *append)
     s->entries = entries;
     s->n += append->n;
 
+done:
     if (append->req->cb != NULL) {
-        append->req->cb(append->req, 0);
+        append->req->cb(append->req, status);
     }
     raft_free(append);
 }
@@ -264,6 +313,7 @@ static void ioFlushSnapshotPut(struct io *s, struct snapshot_put *r)
     if (r->trailing == 0) {
         rv = s->io->truncate(s->io, 1);
         assert(rv == 0);
+        s->start_index = s->snapshot->index;
     }
 
     if (r->req->cb != NULL) {
@@ -286,16 +336,6 @@ static void ioFlushSnapshotGet(struct io *s, struct snapshot_get *r)
     raft_free(r);
 }
 
-/* Flush an async work request */
-static void ioFlushAsyncWork(struct io *s, struct async_work *r)
-{
-    (void) s;
-    int rv;
-    rv = r->req->work(r->req);
-    r->req->cb(r->req, rv);
-    raft_free(r);
-}
-
 /* Search for the peer with the given ID. */
 static struct peer *ioGetPeer(struct io *io, raft_id id)
 {
@@ -348,7 +388,7 @@ static void ioFlushSend(struct io *io, struct send *send)
         goto out;
     }
 
-    transmit = raft_malloc(sizeof *transmit);
+    transmit = raft_calloc(1, sizeof *transmit);
     assert(transmit != NULL);
 
     transmit->type = TRANSMIT;
@@ -361,11 +401,11 @@ static void ioFlushSend(struct io *io, struct send *send)
 
     *dst = *src;
     switch (dst->type) {
-        case RAFT_IO_APPEND_ENTRIES:
+        case RAFT_APPEND_ENTRIES:
             /* Make a copy of the entries being sent */
             copyAppendEntries(&src->append_entries, &dst->append_entries);
             break;
-        case RAFT_IO_INSTALL_SNAPSHOT:
+        case RAFT_INSTALL_SNAPSHOT:
             copyInstallSnapshot(&src->install_snapshot, &dst->install_snapshot);
             break;
     }
@@ -388,13 +428,13 @@ static void ioDestroyTransmit(struct transmit *transmit)
     struct raft_message *message;
     message = &transmit->message;
     switch (message->type) {
-        case RAFT_IO_APPEND_ENTRIES:
+        case RAFT_APPEND_ENTRIES:
             if (message->append_entries.entries != NULL) {
                 raft_free(message->append_entries.entries[0].batch);
                 raft_free(message->append_entries.entries);
             }
             break;
-        case RAFT_IO_INSTALL_SNAPSHOT:
+        case RAFT_INSTALL_SNAPSHOT:
             raft_configuration_close(&message->install_snapshot.conf);
             raft_free(message->install_snapshot.data.base);
             break;
@@ -429,9 +469,6 @@ static void ioFlushAll(struct io *io)
             case SNAPSHOT_GET:
                 ioFlushSnapshotGet(io, (struct snapshot_get *)r);
                 break;
-            case ASYNC_WORK:
-                ioFlushAsyncWork(io, (struct async_work *)r);
-                break;
             default:
                 assert(0);
         }
@@ -600,8 +637,14 @@ static int ioMethodAppend(struct raft_io *raft_io,
 static int ioMethodTruncate(struct raft_io *raft_io, raft_index index)
 {
     struct io *io = raft_io->impl;
+    raft_index last_index = io->start_index + io->n;
     size_t n;
 
+    if (index >= last_index + 1) {
+        /* Nothing to truncate */
+        return 0;
+    }
+
     if (ioFaultTick(io)) {
         return RAFT_IOERR;
     }
@@ -668,28 +711,6 @@ static int ioMethodSnapshotPut(struct raft_io *raft_io,
     return 0;
 }
 
-static int ioMethodAsyncWork(struct raft_io *raft_io,
-                             struct raft_io_async_work *req,
-                             raft_io_async_work_cb cb)
-{
-    struct io *io = raft_io->impl;
-    struct raft *raft = raft_io->data;
-    struct async_work *r;
-
-    r = raft_malloc(sizeof *r);
-    assert(r != NULL);
-
-    r->type = ASYNC_WORK;
-    r->req = req;
-    r->req->cb = cb;
-    r->req->data = raft;
-    r->completion_time = *io->time + io->work_duration;
-
-    QUEUE_PUSH(&io->requests, &r->queue);
-    return 0;
-}
-
-
 static int ioMethodSnapshotGet(struct raft_io *raft_io,
                                struct raft_io_snapshot_get *req,
                                raft_io_snapshot_get_cb cb)
@@ -718,11 +739,14 @@ static raft_time ioMethodTime(struct raft_io *raft_io)
 
 static int ioMethodRandom(struct raft_io *raft_io, int min, int max)
 {
-    struct io *io;
+    (void)raft_io;
     (void)min;
     (void)max;
-    io = raft_io->impl;
-    return (int)io->randomized_election_timeout;
+    /* This method is used by raft_init() to get the initial seed for the
+     * pseudo random number generator. However, it doesn't matter what we return
+     * here, since in serverSeed() we'll call raft_seed() again overwriting
+     * whatever value we return here. */
+    return 0;
 }
 
 /* Queue up a request which will be processed later, when io_stub_flush()
@@ -887,9 +911,11 @@ static int ioInit(struct raft_io *raft_io, unsigned index, raft_time *time)
     io->snapshot = NULL;
     io->entries = NULL;
     io->n = 0;
+    io->start_index = 0;
     QUEUE_INIT(&io->requests);
     io->n_peers = 0;
     io->randomized_election_timeout = ELECTION_TIMEOUT + index * 100;
+    io->randomized_election_timeout_prev = 0;
     io->network_latency = NETWORK_LATENCY;
     io->disk_latency = DISK_LATENCY;
     io->work_duration = WORK_DURATION;
@@ -901,6 +927,8 @@ static int ioInit(struct raft_io *raft_io, unsigned index, raft_time *time)
     io->n_append = 0;
 
     raft_io->impl = io;
+    raft_io->version = 2;
+    raft_io->capacity = 4096;
     raft_io->init = ioMethodInit;
     raft_io->close = ioMethodClose;
     raft_io->start = ioMethodStart;
@@ -913,7 +941,6 @@ static int ioInit(struct raft_io *raft_io, unsigned index, raft_time *time)
     raft_io->truncate = ioMethodTruncate;
     raft_io->send = ioMethodSend;
     raft_io->snapshot_put = ioMethodSnapshotPut;
-    raft_io->async_work = ioMethodAsyncWork;
     raft_io->snapshot_get = ioMethodSnapshotGet;
     raft_io->time = ioMethodTime;
     raft_io->random = ioMethodRandom;
@@ -941,19 +968,32 @@ void ioClose(struct raft_io *raft_io)
 }
 
 /* Custom emit tracer function which include the server ID. */
-static void emit(struct raft_tracer *t,
-                 const char *file,
-                 int line,
-                 const char *message)
+static void fixtureTrace(struct raft_tracer *t, int type, const void *data)
 {
-    unsigned id = *(unsigned *)t->impl;
-    fprintf(stderr, "%d: %30s:%*d - %s\n", id, file, 3, line, message);
+    struct raft_fixture_server *s = t->impl;
+    const struct raft_tracer_info *info = data;
+
+    if (type != RAFT_TRACER_DIAGNOSTIC) {
+        return;
+    }
+
+    if (info->diagnostic.level > 3) {
+        return;
+    }
+
+    fprintf(stderr, "[%4llu] %llu: %s\n", s->io.time(&s->io), s->id,
+            info->diagnostic.message);
 }
 
 static int serverInit(struct raft_fixture *f, unsigned i, struct raft_fsm *fsm)
 {
     int rv;
-    struct raft_fixture_server *s = &f->servers[i];
+    struct raft_fixture_server *s;
+    s = raft_malloc(sizeof(*s));
+    if (s == NULL) {
+        return RAFT_NOMEM;
+    }
+    f->servers[i] = s;
     s->alive = true;
     s->id = i + 1;
     sprintf(s->address, "%llu", s->id);
@@ -968,10 +1008,10 @@ static int serverInit(struct raft_fixture *f, unsigned i, struct raft_fsm *fsm)
     raft_set_election_timeout(&s->raft, ELECTION_TIMEOUT);
     raft_set_heartbeat_timeout(&s->raft, HEARTBEAT_TIMEOUT);
     raft_set_install_snapshot_timeout(&s->raft, INSTALL_SNAPSHOT_TIMEOUT);
-    s->tracer.impl = (void *)&s->id;
-    s->tracer.emit = emit;
+    s->tracer.impl = s;
+    s->tracer.version = 2;
+    s->tracer.emit = fixtureTrace;
     s->raft.tracer = &s->tracer;
-    raft_tracer_maybe_enable(&s->tracer, true);
     return 0;
 }
 
@@ -979,6 +1019,35 @@ static void serverClose(struct raft_fixture_server *s)
 {
     raft_close(&s->raft, NULL);
     ioClose(&s->io);
+    raft_free(s);
+}
+
+/* Set the state of raft's internal pseudo random number generator so that the
+ * next time RandomWithinRange() is run it will return the value configured by
+ * the fixture, which is stored in the randomized_election_timeout field. */
+static void serverSeed(struct raft_fixture_server *s)
+{
+    struct io *io = s->io.impl;
+    unsigned timeout = s->raft.election_timeout;
+
+    if (io->randomized_election_timeout ==
+        io->randomized_election_timeout_prev) {
+        goto done;
+    }
+
+    io->seed = s->raft.random;
+    while (1) {
+        unsigned random = io->seed;
+        unsigned n = RandomWithinRange(&random, timeout, timeout * 2);
+        if (n == io->randomized_election_timeout) {
+            goto done;
+        }
+        io->seed = random;
+    }
+
+done:
+    raft_seed(&s->raft, io->seed);
+    io->randomized_election_timeout_prev = io->randomized_election_timeout;
 }
 
 /* Connect the server with the given index to all others */
@@ -986,8 +1055,8 @@ static void serverConnectToAll(struct raft_fixture *f, unsigned i)
 {
     unsigned j;
     for (j = 0; j < f->n; j++) {
-        struct raft_io *io1 = &f->servers[i].io;
-        struct raft_io *io2 = &f->servers[j].io;
+        struct raft_io *io1 = &f->servers[i]->io;
+        struct raft_io *io2 = &f->servers[j]->io;
         if (i == j) {
             continue;
         }
@@ -995,55 +1064,40 @@ static void serverConnectToAll(struct raft_fixture *f, unsigned i)
     }
 }
 
-int raft_fixture_init(struct raft_fixture *f, unsigned n, struct raft_fsm *fsms)
+int raft_fixture_init(struct raft_fixture *f)
 {
-    unsigned i;
-    int rc;
-    assert(n >= 1);
-
     f->time = 0;
-    f->n = n;
-
-    /* Initialize all servers */
-    for (i = 0; i < n; i++) {
-        rc = serverInit(f, i, &fsms[i]);
-        if (rc != 0) {
-            return rc;
-        }
+    f->n = 0;
+    f->log = logInit();
+    if (f->log == NULL) {
+        return RAFT_NOMEM;
     }
-
-    /* Connect all servers to each another */
-    for (i = 0; i < f->n; i++) {
-        serverConnectToAll(f, i);
-    }
-
-    logInit(&f->log);
     f->commit_index = 0;
     f->hook = NULL;
-
+    f->event = raft_malloc(sizeof(*f->event));
+    if (f->event == NULL) {
+        return RAFT_NOMEM;
+    }
     return 0;
 }
 
 int raft_fixture_initialize(struct raft_fixture *f)
 {
-    f->time = 0;
-    logInit(&f->log);
-    f->commit_index = 0;
-    f->hook = NULL;
-    return 0;
+    return raft_fixture_init(f);
 }
 
 void raft_fixture_close(struct raft_fixture *f)
 {
     unsigned i;
     for (i = 0; i < f->n; i++) {
-        struct io *io = f->servers[i].io.impl;
+        struct io *io = f->servers[i]->io.impl;
         ioFlushAll(io);
     }
     for (i = 0; i < f->n; i++) {
-        serverClose(&f->servers[i]);
+        serverClose(f->servers[i]);
     }
-    logClose(&f->log);
+    raft_free(f->event);
+    logClose(f->log);
 }
 
 int raft_fixture_configuration(struct raft_fixture *f,
@@ -1059,7 +1113,7 @@ int raft_fixture_configuration(struct raft_fixture *f,
         struct raft_fixture_server *s;
         int role = i < n_voting ? RAFT_VOTER : RAFT_STANDBY;
         int rv;
-        s = &f->servers[i];
+        s = f->servers[i];
         rv = raft_configuration_add(configuration, s->id, s->address, role);
         if (rv != 0) {
             return rv;
@@ -1083,12 +1137,21 @@ int raft_fixture_bootstrap(struct raft_fixture *f,
     return 0;
 }
 
+static void seedAll(struct raft_fixture *f)
+{
+    unsigned i;
+    for (i = 0; i < f->n; i++) {
+        serverSeed(f->servers[i]);
+    }
+}
+
 int raft_fixture_start(struct raft_fixture *f)
 {
     unsigned i;
     int rv;
+    seedAll(f);
     for (i = 0; i < f->n; i++) {
-        struct raft_fixture_server *s = &f->servers[i];
+        struct raft_fixture_server *s = f->servers[i];
         rv = raft_start(&s->raft);
         if (rv != 0) {
             return rv;
@@ -1110,13 +1173,13 @@ raft_time raft_fixture_time(struct raft_fixture *f)
 struct raft *raft_fixture_get(struct raft_fixture *f, unsigned i)
 {
     assert(i < f->n);
-    return &f->servers[i].raft;
+    return &f->servers[i]->raft;
 }
 
 bool raft_fixture_alive(struct raft_fixture *f, unsigned i)
 {
     assert(i < f->n);
-    return f->servers[i].alive;
+    return f->servers[i]->alive;
 }
 
 unsigned raft_fixture_leader_index(struct raft_fixture *f)
@@ -1129,7 +1192,7 @@ unsigned raft_fixture_leader_index(struct raft_fixture *f)
 
 raft_id raft_fixture_voted_for(struct raft_fixture *f, unsigned i)
 {
-    struct io *io = f->servers[i].io.impl;
+    struct io *io = f->servers[i]->io.impl;
     return io->voted_for;
 }
 
@@ -1258,7 +1321,7 @@ static void checkLeaderAppendOnly(struct raft_fixture *f)
 {
     struct raft *raft;
     raft_index index;
-    raft_index last = logLastIndex(&f->log);
+    raft_index last = logLastIndex(f->log);
 
     /* If the cached log is empty it means there was no leader before. */
     if (last == 0) {
@@ -1271,21 +1334,21 @@ static void checkLeaderAppendOnly(struct raft_fixture *f)
     }
 
     raft = raft_fixture_get(f, (unsigned)f->leader_id - 1);
-    last = logLastIndex(&f->log);
+    last = logLastIndex(f->log);
 
     for (index = 1; index <= last; index++) {
         const struct raft_entry *entry1;
         const struct raft_entry *entry2;
         size_t i;
 
-        entry1 = logGet(&f->log, index);
-        entry2 = logGet(&raft->log, index);
+        entry1 = logGet(f->log, index);
+        entry2 = logGet(raft->legacy.log, index);
 
         assert(entry1 != NULL);
 
         /* Check if the entry was snapshotted. */
         if (entry2 == NULL) {
-            assert(raft->log.snapshot.last_index >= index);
+            assert(raft->legacy.log->snapshot.last_index >= index);
             continue;
         }
 
@@ -1308,9 +1371,14 @@ static void copyLeaderLog(struct raft_fixture *f)
     unsigned n;
     size_t i;
     int rv;
-    logClose(&f->log);
-    logInit(&f->log);
-    rv = logAcquire(&raft->log, 1, &entries, &n);
+    logClose(f->log);
+    f->log = logInit();
+    if (f->log == NULL) {
+        assert(false);
+        return;
+    }
+
+    rv = logAcquire(raft->legacy.log, 1, &entries, &n);
     assert(rv == 0);
     for (i = 0; i < n; i++) {
         struct raft_entry *entry = &entries[i];
@@ -1319,10 +1387,10 @@ static void copyLeaderLog(struct raft_fixture *f)
         buf.base = raft_malloc(buf.len);
         assert(buf.base != NULL);
         memcpy(buf.base, entry->buf.base, buf.len);
-        rv = logAppend(&f->log, entry->term, entry->type, &buf, NULL);
+        rv = logAppend(f->log, entry->term, entry->type, &buf, NULL);
         assert(rv == 0);
     }
-    logRelease(&raft->log, 1, entries, n);
+    logRelease(raft->legacy.log, 1, entries, n);
 }
 
 /* Update the commit index to match the one from the current leader. */
@@ -1341,7 +1409,7 @@ static void getLowestTickTime(struct raft_fixture *f, raft_time *t, unsigned *i)
     unsigned j;
     *t = (raft_time)-1 /* Maximum value */;
     for (j = 0; j < f->n; j++) {
-        struct io *io = f->servers[j].io.impl;
+        struct io *io = f->servers[j]->io.impl;
         if (io->next_tick < *t) {
             *t = io->next_tick;
             *i = j;
@@ -1358,10 +1426,9 @@ static void getLowestRequestCompletionTime(struct raft_fixture *f,
     unsigned j;
     *t = (raft_time)-1 /* Maximum value */;
     for (j = 0; j < f->n; j++) {
-        struct io *io = f->servers[j].io.impl;
+        struct io *io = f->servers[j]->io.impl;
         queue *head;
-        QUEUE_FOREACH(head, &io->requests)
-        {
+        QUEUE_FOREACH (head, &io->requests) {
             struct ioRequest *r = QUEUE_DATA(head, struct ioRequest, queue);
             if (r->completion_time < *t) {
                 *t = r->completion_time;
@@ -1374,12 +1441,12 @@ static void getLowestRequestCompletionTime(struct raft_fixture *f,
 /* Fire the tick callback of the i'th server. */
 static void fireTick(struct raft_fixture *f, unsigned i)
 {
-    struct io *io = f->servers[i].io.impl;
+    struct io *io = f->servers[i]->io.impl;
     f->time = io->next_tick;
-    f->event.server_index = i;
-    f->event.type = RAFT_FIXTURE_TICK;
+    f->event->server_index = i;
+    f->event->type = RAFT_FIXTURE_TICK;
     io->next_tick += io->tick_interval;
-    if (f->servers[i].alive) {
+    if (f->servers[i]->alive) {
         io->tick_cb(io->io);
     }
 }
@@ -1387,14 +1454,13 @@ static void fireTick(struct raft_fixture *f, unsigned i)
 /* Complete the first request with completion time @t on the @i'th server. */
 static void completeRequest(struct raft_fixture *f, unsigned i, raft_time t)
 {
-    struct io *io = f->servers[i].io.impl;
+    struct io *io = f->servers[i]->io.impl;
     queue *head;
     struct ioRequest *r = NULL;
     bool found = false;
     f->time = t;
-    f->event.server_index = i;
-    QUEUE_FOREACH(head, &io->requests)
-    {
+    f->event->server_index = i;
+    QUEUE_FOREACH (head, &io->requests) {
         r = QUEUE_DATA(head, struct ioRequest, queue);
         if (r->completion_time == t) {
             found = true;
@@ -1406,27 +1472,23 @@ static void completeRequest(struct raft_fixture *f, unsigned i, raft_time t)
     switch (r->type) {
         case APPEND:
             ioFlushAppend(io, (struct append *)r);
-            f->event.type = RAFT_FIXTURE_DISK;
+            f->event->type = RAFT_FIXTURE_DISK;
             break;
         case SEND:
             ioFlushSend(io, (struct send *)r);
-            f->event.type = RAFT_FIXTURE_NETWORK;
+            f->event->type = RAFT_FIXTURE_NETWORK;
             break;
         case TRANSMIT:
             ioDeliverTransmit(io, (struct transmit *)r);
-            f->event.type = RAFT_FIXTURE_NETWORK;
+            f->event->type = RAFT_FIXTURE_NETWORK;
             break;
         case SNAPSHOT_PUT:
             ioFlushSnapshotPut(io, (struct snapshot_put *)r);
-            f->event.type = RAFT_FIXTURE_DISK;
+            f->event->type = RAFT_FIXTURE_DISK;
             break;
         case SNAPSHOT_GET:
             ioFlushSnapshotGet(io, (struct snapshot_get *)r);
-            f->event.type = RAFT_FIXTURE_DISK;
-            break;
-        case ASYNC_WORK:
-            ioFlushAsyncWork(io, (struct async_work *)r);
-            f->event.type = RAFT_FIXTURE_WORK;
+            f->event->type = RAFT_FIXTURE_DISK;
             break;
         default:
             assert(0);
@@ -1440,9 +1502,10 @@ struct raft_fixture_event *raft_fixture_step(struct raft_fixture *f)
     unsigned i = f->n;
     unsigned j = f->n;
 
+    seedAll(f);
+
     getLowestTickTime(f, &tick_time, &i);
     getLowestRequestCompletionTime(f, &completion_time, &j);
-
     assert(i < f->n || j < f->n);
 
     if (tick_time < completion_time ||
@@ -1452,6 +1515,11 @@ struct raft_fixture_event *raft_fixture_step(struct raft_fixture *f)
         completeRequest(f, j, completion_time);
     }
 
+    for (j = 0; j < f->n; j++) {
+        struct raft *r = &f->servers[j]->raft;
+        LegacyFireCompletedRequests(r);
+    }
+
     /* If the leader has not changed check the Leader Append-Only
      * guarantee. */
     if (!updateLeaderAndCheckElectionSafety(f)) {
@@ -1465,10 +1533,10 @@ struct raft_fixture_event *raft_fixture_step(struct raft_fixture *f)
     }
 
     if (f->hook != NULL) {
-        f->hook(f, &f->event);
+        f->hook(f, f->event);
     }
 
-    return &f->event;
+    return f->event;
 }
 
 struct raft_fixture_event *raft_fixture_step_n(struct raft_fixture *f,
@@ -1494,8 +1562,8 @@ bool raft_fixture_step_until(struct raft_fixture *f,
     return f->time - start < max_msecs;
 }
 
-/* A step function which return always false, forcing raft_fixture_step_n to
- * advance time at each iteration. */
+/* A step function which return always false, forcing
+ * raft_fixture_step_n to advance time at each iteration. */
 static bool spin(struct raft_fixture *f, void *arg)
 {
     (void)f;
@@ -1532,8 +1600,8 @@ bool raft_fixture_step_until_has_no_leader(struct raft_fixture *f,
     return raft_fixture_step_until(f, hasNoLeader, NULL, max_msecs);
 }
 
-/* Enable/disable dropping outgoing messages of a certain type from all servers
- * except one. */
+/* Enable/disable dropping outgoing messages of a certain type from all
+ * servers except one. */
 static void dropAllExcept(struct raft_fixture *f,
                           int type,
                           bool flag,
@@ -1541,7 +1609,7 @@ static void dropAllExcept(struct raft_fixture *f,
 {
     unsigned j;
     for (j = 0; j < f->n; j++) {
-        struct raft_fixture_server *s = &f->servers[j];
+        struct raft_fixture_server *s = f->servers[j];
         if (j == i) {
             continue;
         }
@@ -1549,18 +1617,18 @@ static void dropAllExcept(struct raft_fixture *f,
     }
 }
 
-/* Set the randomized election timeout of the given server to the minimum value
- * compatible with its current state and timers. */
+/* Set the randomized election timeout of the given server to the
+ * minimum value compatible with its current state and timers. */
 static void minimizeRandomizedElectionTimeout(struct raft_fixture *f,
                                               unsigned i)
 {
-    struct raft *raft = &f->servers[i].raft;
+    struct raft *raft = &f->servers[i]->raft;
     raft_time now = raft->io->time(raft->io);
     unsigned timeout = raft->election_timeout;
     assert(raft->state == RAFT_FOLLOWER);
 
-    /* If the minimum election timeout value would make the timer expire in the
-     * past, cap it. */
+    /* If the minimum election timeout value would make the timer expire
+     * in the past, cap it. */
     if (now - raft->election_timer_start > timeout) {
         timeout = (unsigned)(now - raft->election_timer_start);
     }
@@ -1568,14 +1636,14 @@ static void minimizeRandomizedElectionTimeout(struct raft_fixture *f,
     raft->follower_state.randomized_election_timeout = timeout;
 }
 
-/* Set the randomized election timeout to the maximum value on all servers
- * except the given one. */
+/* Set the randomized election timeout to the maximum value on all
+ * servers except the given one. */
 static void maximizeAllRandomizedElectionTimeoutsExcept(struct raft_fixture *f,
                                                         unsigned i)
 {
     unsigned j;
     for (j = 0; j < f->n; j++) {
-        struct raft *raft = &f->servers[j].raft;
+        struct raft *raft = &f->servers[j]->raft;
         unsigned timeout = raft->election_timeout * 2;
         if (j == i) {
             continue;
@@ -1590,7 +1658,7 @@ void raft_fixture_hook(struct raft_fixture *f, raft_fixture_event_cb hook)
     f->hook = hook;
 }
 
-void raft_fixture_elect(struct raft_fixture *f, unsigned i)
+void raft_fixture_start_elect(struct raft_fixture *f, unsigned i)
 {
     struct raft *raft = raft_fixture_get(f, i);
     unsigned j;
@@ -1604,15 +1672,21 @@ void raft_fixture_elect(struct raft_fixture *f, unsigned i)
 
     /* Make sure all servers are currently followers. */
     for (j = 0; j < f->n; j++) {
-        assert(raft_state(&f->servers[j].raft) == RAFT_FOLLOWER);
+        assert(raft_state(&f->servers[j]->raft) == RAFT_FOLLOWER);
     }
 
-    /* Pretend that the last randomized election timeout was set at the maximum
-     * value on all server expect the one to be elected, which is instead set to
-     * the minimum possible value compatible with its current state. */
+    /* Pretend that the last randomized election timeout was set at the
+     * maximum value on all server expect the one to be elected, which
+     * is instead set to the minimum possible value compatible with its
+     * current state. */
     minimizeRandomizedElectionTimeout(f, i);
     maximizeAllRandomizedElectionTimeoutsExcept(f, i);
+}
 
+void raft_fixture_elect(struct raft_fixture *f, unsigned i)
+{
+    struct raft *raft = raft_fixture_get(f, i);
+    raft_fixture_start_elect(f, i);
     raft_fixture_step_until_has_leader(f, ELECTION_TIMEOUT * 20);
     assert(f->leader_id == raft->id);
 }
@@ -1624,20 +1698,20 @@ void raft_fixture_depose(struct raft_fixture *f)
     /* Make sure there's a leader. */
     assert(f->leader_id != 0);
     leader_i = (unsigned)f->leader_id - 1;
-    assert(raft_state(&f->servers[leader_i].raft) == RAFT_LEADER);
+    assert(raft_state(&f->servers[leader_i]->raft) == RAFT_LEADER);
 
-    /* Set a very large election timeout on all followers, to prevent them from
-     * starting an election. */
+    /* Set a very large election timeout on all followers, to prevent
+     * them from starting an election. */
     maximizeAllRandomizedElectionTimeoutsExcept(f, leader_i);
 
-    /* Prevent all servers from sending append entries results, so the leader
-     * will eventually step down. */
-    dropAllExcept(f, RAFT_IO_APPEND_ENTRIES_RESULT, true, leader_i);
+    /* Prevent all servers from sending append entries results, so the
+     * leader will eventually step down. */
+    dropAllExcept(f, RAFT_APPEND_ENTRIES_RESULT, true, leader_i);
 
     raft_fixture_step_until_has_no_leader(f, ELECTION_TIMEOUT * 3);
     assert(f->leader_id == 0);
 
-    dropAllExcept(f, RAFT_IO_APPEND_ENTRIES_RESULT, false, leader_i);
+    dropAllExcept(f, RAFT_APPEND_ENTRIES_RESULT, false, leader_i);
 }
 
 struct step_apply
@@ -1664,6 +1738,7 @@ static bool hasAppliedIndex(struct raft_fixture *f, void *arg)
             n++;
         }
     }
+
     return n == f->n;
 }
 
@@ -1760,8 +1835,7 @@ static bool hasDelivered(struct raft_fixture *f, void *arg)
     queue *head;
     raft = raft_fixture_get(f, target->i);
     io = raft->io->impl;
-    QUEUE_FOREACH(head, &io->requests)
-    {
+    QUEUE_FOREACH (head, &io->requests) {
         struct ioRequest *r;
         r = QUEUE_DATA(head, struct ioRequest, queue);
         message = NULL;
@@ -1791,22 +1865,22 @@ bool raft_fixture_step_until_delivered(struct raft_fixture *f,
 
 void raft_fixture_disconnect(struct raft_fixture *f, unsigned i, unsigned j)
 {
-    struct raft_io *io1 = &f->servers[i].io;
-    struct raft_io *io2 = &f->servers[j].io;
+    struct raft_io *io1 = &f->servers[i]->io;
+    struct raft_io *io2 = &f->servers[j]->io;
     ioDisconnect(io1, io2);
 }
 
 void raft_fixture_reconnect(struct raft_fixture *f, unsigned i, unsigned j)
 {
-    struct raft_io *io1 = &f->servers[i].io;
-    struct raft_io *io2 = &f->servers[j].io;
+    struct raft_io *io1 = &f->servers[i]->io;
+    struct raft_io *io2 = &f->servers[j]->io;
     ioReconnect(io1, io2);
 }
 
 void raft_fixture_saturate(struct raft_fixture *f, unsigned i, unsigned j)
 {
-    struct raft_io *io1 = &f->servers[i].io;
-    struct raft_io *io2 = &f->servers[j].io;
+    struct raft_io *io1 = &f->servers[i]->io;
+    struct raft_io *io2 = &f->servers[j]->io;
     ioSaturate(io1, io2);
 }
 
@@ -1830,7 +1904,7 @@ static void reconnectToAll(struct raft_fixture *f, unsigned i)
             continue;
         }
         /* Don't reconnect to disconnected peers */
-        if (!f->servers[j].alive) {
+        if (!f->servers[j]->alive) {
             continue;
         }
         raft_fixture_desaturate(f, i, j);
@@ -1840,28 +1914,28 @@ static void reconnectToAll(struct raft_fixture *f, unsigned i)
 
 bool raft_fixture_saturated(struct raft_fixture *f, unsigned i, unsigned j)
 {
-    struct raft_io *io1 = &f->servers[i].io;
-    struct raft_io *io2 = &f->servers[j].io;
+    struct raft_io *io1 = &f->servers[i]->io;
+    struct raft_io *io2 = &f->servers[j]->io;
     return ioSaturated(io1, io2);
 }
 
 void raft_fixture_desaturate(struct raft_fixture *f, unsigned i, unsigned j)
 {
-    struct raft_io *io1 = &f->servers[i].io;
-    struct raft_io *io2 = &f->servers[j].io;
+    struct raft_io *io1 = &f->servers[i]->io;
+    struct raft_io *io2 = &f->servers[j]->io;
     ioDesaturate(io1, io2);
 }
 
 void raft_fixture_kill(struct raft_fixture *f, unsigned i)
 {
     disconnectFromAll(f, i);
-    f->servers[i].alive = false;
+    f->servers[i]->alive = false;
 }
 
 void raft_fixture_revive(struct raft_fixture *f, unsigned i)
 {
     reconnectToAll(f, i);
-    f->servers[i].alive = true;
+    f->servers[i]->alive = true;
 }
 
 int raft_fixture_grow(struct raft_fixture *f, struct raft_fsm *fsm)
@@ -1879,8 +1953,8 @@ int raft_fixture_grow(struct raft_fixture *f, struct raft_fsm *fsm)
 
     serverConnectToAll(f, i);
     for (j = 0; j < f->n; j++) {
-        struct raft_io *io1 = &f->servers[i].io;
-        struct raft_io *io2 = &f->servers[j].io;
+        struct raft_io *io1 = &f->servers[i]->io;
+        struct raft_io *io2 = &f->servers[j]->io;
         ioConnect(io2, io1);
     }
 
@@ -1891,7 +1965,7 @@ void raft_fixture_set_randomized_election_timeout(struct raft_fixture *f,
                                                   unsigned i,
                                                   unsigned msecs)
 {
-    struct io *io = f->servers[i].io.impl;
+    struct io *io = f->servers[i]->io.impl;
     io->randomized_election_timeout = msecs;
 }
 
@@ -1899,7 +1973,7 @@ void raft_fixture_set_network_latency(struct raft_fixture *f,
                                       unsigned i,
                                       unsigned msecs)
 {
-    struct io *io = f->servers[i].io.impl;
+    struct io *io = f->servers[i]->io.impl;
     io->network_latency = msecs;
 }
 
@@ -1907,13 +1981,13 @@ void raft_fixture_set_disk_latency(struct raft_fixture *f,
                                    unsigned i,
                                    unsigned msecs)
 {
-    struct io *io = f->servers[i].io.impl;
+    struct io *io = f->servers[i]->io.impl;
     io->disk_latency = msecs;
 }
 
 void raft_fixture_set_term(struct raft_fixture *f, unsigned i, raft_term term)
 {
-    struct io *io = f->servers[i].io.impl;
+    struct io *io = f->servers[i]->io.impl;
     io->term = term;
 }
 
@@ -1921,7 +1995,7 @@ void raft_fixture_set_snapshot(struct raft_fixture *f,
                                unsigned i,
                                struct raft_snapshot *snapshot)
 {
-    struct io *io = f->servers[i].io.impl;
+    struct io *io = f->servers[i]->io.impl;
     io->snapshot = snapshot;
 }
 
@@ -1929,7 +2003,7 @@ void raft_fixture_add_entry(struct raft_fixture *f,
                             unsigned i,
                             struct raft_entry *entry)
 {
-    struct io *io = f->servers[i].io.impl;
+    struct io *io = f->servers[i]->io.impl;
     struct raft_entry *entries;
     entries = raft_realloc(io->entries, (io->n + 1) * sizeof *entries);
     assert(entries != NULL);
@@ -1943,20 +2017,20 @@ void raft_fixture_io_fault(struct raft_fixture *f,
                            int delay,
                            int repeat)
 {
-    struct io *io = f->servers[i].io.impl;
+    struct io *io = f->servers[i]->io.impl;
     io->fault.countdown = delay;
     io->fault.n = repeat;
 }
 
 unsigned raft_fixture_n_send(struct raft_fixture *f, unsigned i, int type)
 {
-    struct io *io = f->servers[i].io.impl;
+    struct io *io = f->servers[i]->io.impl;
     return io->n_send[type];
 }
 
 unsigned raft_fixture_n_recv(struct raft_fixture *f, unsigned i, int type)
 {
-    struct io *io = f->servers[i].io.impl;
+    struct io *io = f->servers[i]->io.impl;
     return io->n_recv[type];
 }
 
