@@ -81,10 +81,17 @@ static int refsTryInsert(struct raft_entry_ref *table,
         /* All entries in a bucket must have the same index. */
         assert(next_slot->index == index);
 
-        /* It should never happen that two entries with the same index and term
-         * get appended. So no existing slot in this bucket must track an entry
-         * with the same term as the given one. */
-        assert(next_slot->term != term);
+        /* It might happen that two entries with the same index and term get
+         * appended. For example if the same entry was truncated by a leader
+         * because it failed to be written to disk, but it is then received
+         * again from a new leader, while the old leader's reference count for
+         * that entry hasn't dropped to zero yet (e.g. it's being sent). In that
+         * case we return an error, and let the leader retry the AppendEntries,
+         * until it eventually succeed when all references to the old entry are
+         * gone. */
+        if (next_slot->term == term) {
+            return RAFT_BUSY;
+        }
 
         last_slot = next_slot;
     }
@@ -235,7 +242,7 @@ static int refsInit(struct raft_log *l,
 
         rc = refsTryInsert(l->refs, l->refs_size, term, index, 1, &collision);
         if (rc != 0) {
-            return RAFT_NOMEM;
+            return rc;
         }
 
         if (!collision) {
@@ -335,17 +342,25 @@ static bool refsDecr(struct raft_log *l,
     return true;
 }
 
-void logInit(struct raft_log *l)
+struct raft_log *logInit(void)
 {
-    assert(l != NULL);
-    l->entries = NULL;
-    l->size = 0;
-    l->front = l->back = 0;
-    l->offset = 0;
-    l->refs = NULL;
-    l->refs_size = 0;
-    l->snapshot.last_index = 0;
-    l->snapshot.last_term = 0;
+    struct raft_log *log;
+
+    log = raft_malloc(sizeof(*log));
+    if (log == NULL) {
+        return NULL;
+    }
+
+    log->entries = NULL;
+    log->size = 0;
+    log->front = log->back = 0;
+    log->offset = 0;
+    log->refs = NULL;
+    log->refs_size = 0;
+    log->snapshot.last_index = 0;
+    log->snapshot.last_term = 0;
+
+    return log;
 }
 
 /* Return the index of the i'th entry in the log. */
@@ -411,6 +426,8 @@ void logClose(struct raft_log *l)
     if (l->refs != NULL) {
         raft_free(l->refs);
     }
+
+    raft_free(l);
 }
 
 void logStart(struct raft_log *l,
@@ -427,7 +444,8 @@ void logStart(struct raft_log *l,
     l->offset = start_index - 1;
 }
 
-/* Ensure that the entries array has enough free slots for adding a new entry. */
+/* Ensure that the entries array has enough free slots for adding a new entry.
+ */
 static int ensureCapacity(struct raft_log *l)
 {
     struct raft_entry *entries; /* New entries array */
@@ -486,14 +504,16 @@ int logAppend(struct raft_log *l,
 
     rv = ensureCapacity(l);
     if (rv != 0) {
-        return rv;
+        assert(rv == RAFT_NOMEM);
+        goto err;
     }
 
     index = logLastIndex(l) + 1;
 
     rv = refsInit(l, term, index);
     if (rv != 0) {
-        return rv;
+        assert(rv == RAFT_NOMEM || rv == RAFT_BUSY);
+        goto err;
     }
 
     entry = &l->entries[l->back];
@@ -506,6 +526,10 @@ int logAppend(struct raft_log *l,
     l->back = l->back % l->size;
 
     return 0;
+
+err:
+    assert(rv == RAFT_NOMEM || rv == RAFT_BUSY);
+    return rv;
 }
 
 int logAppendCommands(struct raft_log *l,
@@ -619,8 +643,8 @@ raft_term logTermOf(struct raft_log *l, const raft_index index)
 
     if (index == l->snapshot.last_index) {
         assert(l->snapshot.last_term != 0);
-        /* Coherence check that if we still have the entry at last_index, its term
-         * matches the one in the snapshot. */
+        /* Coherence check that if we still have the entry at last_index, its
+         * term matches the one in the snapshot. */
         i = locateEntry(l, index);
         if (i != l->size) {
             assert(l->entries[i].term == l->snapshot.last_term);
@@ -638,6 +662,11 @@ raft_index logSnapshotIndex(struct raft_log *l)
     return l->snapshot.last_index;
 }
 
+raft_term logSnapshotTerm(struct raft_log *l)
+{
+    return l->snapshot.last_term;
+}
+
 raft_term logLastTerm(struct raft_log *l)
 {
     raft_index last_index;
@@ -662,10 +691,11 @@ const struct raft_entry *logGet(struct raft_log *l, const raft_index index)
     return &l->entries[i];
 }
 
-int logAcquire(struct raft_log *l,
-               const raft_index index,
-               struct raft_entry *entries[],
-               unsigned *n)
+int logAcquireAtMost(struct raft_log *l,
+                     const raft_index index,
+                     int max,
+                     struct raft_entry *entries[],
+                     unsigned *n)
 {
     size_t i;
     size_t j;
@@ -678,7 +708,7 @@ int logAcquire(struct raft_log *l,
     /* Get the array index of the first entry to acquire. */
     i = locateEntry(l, index);
 
-    if (i == l->size) {
+    if (i == l->size || max == 0) {
         *n = 0;
         *entries = NULL;
         return 0;
@@ -697,6 +727,10 @@ int logAcquire(struct raft_log *l,
 
     assert(*n > 0);
 
+    if (max != -1 && *n > (unsigned)max) {
+        *n = (unsigned)max;
+    }
+
     *entries = raft_calloc(*n, sizeof **entries);
     if (*entries == NULL) {
         return RAFT_NOMEM;
@@ -712,6 +746,14 @@ int logAcquire(struct raft_log *l,
     return 0;
 }
 
+int logAcquire(struct raft_log *l,
+               raft_index index,
+               struct raft_entry *entries[],
+               unsigned *n)
+{
+    return logAcquireAtMost(l, index, -1, entries, n);
+}
+
 /* Return true if the given batch is referenced by any entry currently in the
  * log. */
 static bool isBatchReferenced(struct raft_log *l, const void *batch)
