@@ -1,8 +1,8 @@
 #include "assert.h"
 #include "configuration.h"
 #include "election.h"
-#include "log.h"
 #include "queue.h"
+#include "trail.h"
 
 int raft_state(struct raft *r)
 {
@@ -12,7 +12,6 @@ int raft_state(struct raft *r)
 void raft_leader(struct raft *r, raft_id *id, const char **address)
 {
     switch (r->state) {
-        case RAFT_UNAVAILABLE:
         case RAFT_CANDIDATE:
             *id = 0;
             *address = NULL;
@@ -22,7 +21,7 @@ void raft_leader(struct raft *r, raft_id *id, const char **address)
             *address = r->follower_state.current_leader.address;
             return;
         case RAFT_LEADER:
-            if (r->transfer != NULL) {
+            if (r->leader_state.transferee != 0) {
                 *id = 0;
                 *address = NULL;
                 return;
@@ -35,10 +34,15 @@ void raft_leader(struct raft *r, raft_id *id, const char **address)
 
 raft_index raft_last_index(struct raft *r)
 {
-    return logLastIndex(&r->log);
+    return TrailLastIndex(&r->trail);
 }
 
-raft_index raft_last_applied(struct raft *r)
+int raft_role(struct raft *r)
 {
-    return r->last_applied;
+    const struct raft_server *local =
+        configurationGet(&r->configuration, r->id);
+    if (local == NULL) {
+        return -1;
+    }
+    return local->role;
 }
