@@ -6,6 +6,8 @@
 #include "uv_encoding.h"
 #include "uv_writer.h"
 
+#define tracef(...) Tracef(uv->tracer, __VA_ARGS__)
+
 /* The happy path for an append request is:
  *
  * - If there is a current segment and it is has enough spare capacity to hold
@@ -72,10 +74,8 @@ struct uvAppend
 static void uvAliveSegmentWriterCloseCb(struct UvWriter *writer)
 {
     struct uvAliveSegment *segment = writer->data;
-    struct uv *uv = segment->uv;
     uvSegmentBufferClose(&segment->pending);
     RaftHeapFree(segment);
-    uvMaybeFireCloseCb(uv);
 }
 
 /* Submit a request to close the current open segment. */
@@ -108,7 +108,11 @@ static void uvAppendFinishRequestsInQueue(struct uv *uv, queue *q, int status)
         append = QUEUE_DATA(head, struct uvAppend, queue);
         /* Rollback the append next index if the result was unsuccessful. */
         if (status != 0) {
+            tracef("rollback uv->append_next_index was:%llu",
+                   uv->append_next_index);
             uv->append_next_index -= append->n;
+            tracef("rollback uv->append_next_index now:%llu",
+                   uv->append_next_index);
         }
         QUEUE_REMOVE(head);
         QUEUE_PUSH(&queue_copy, head);
@@ -179,7 +183,29 @@ static int uvAliveSegmentEncodeEntriesToWriteBuf(struct uvAliveSegment *segment,
     return 0;
 }
 
+static void uvAliveSegmentWriteCb(struct UvWriterReq *write, const int status);
+
+static void uvAppendRetryTimerCb(uv_timer_t *timer)
+{
+    struct uvAliveSegment *s = timer->data;
+    struct uv *uv = s->uv;
+    int rv;
+
+    uv->append_retry.data = uv;
+
+    rv = UvWriterSubmit(&s->writer, &s->write, &s->buf, 1,
+                        s->next_block * s->uv->block_size,
+                        uvAliveSegmentWriteCb);
+    if (rv != 0) {
+        uv->append_retry.data = s;
+        rv = uv_timer_start(&uv->prepare_retry, uvAppendRetryTimerCb,
+                            uv->disk_retry, 0);
+        assert(rv == 0);
+    }
+}
+
 static int uvAppendMaybeStart(struct uv *uv);
+
 static void uvAliveSegmentWriteCb(struct UvWriterReq *write, const int status)
 {
     struct uvAliveSegment *s = write->data;
@@ -192,11 +218,14 @@ static void uvAliveSegmentWriteCb(struct UvWriterReq *write, const int status)
     assert(s->buf.len % uv->block_size == 0);
     assert(s->buf.len >= uv->block_size);
 
-    /* Check if the write was successful. */
+    /* If the write was unsuccessful, retry it after a delay. */
     if (status != 0) {
-        Tracef(uv->tracer, "write: %s", uv->io->errmsg);
-        uv->errored = true;
-        goto out;
+        Tracef(uv->tracer, "retry failed write (%s)", uv->io->errmsg);
+        uv->append_retry.data = s;
+        rv = uv_timer_start(&uv->prepare_retry, uvAppendRetryTimerCb,
+                            uv->disk_retry, 0);
+        assert(rv == 0);
+        return;
     }
 
     s->written = s->next_block * uv->block_size + s->pending.n;
@@ -250,7 +279,6 @@ static void uvAliveSegmentWriteCb(struct UvWriterReq *write, const int status)
         }
     }
 
-out:
     /* Fire the callbacks of all requests that were fulfilled with this
      * write. */
     uvAppendFinishWritingRequests(uv, status);
@@ -270,14 +298,14 @@ out:
         if (rv != 0) {
             uv->errored = true;
         }
-    } else if (s->finalize
-               && (s->pending_last_index == s->last_index)
-               && !s->writer.closing) {
+    } else if (s->finalize && (s->pending_last_index == s->last_index) &&
+               !s->writer.closing) {
         /* If there are no more append_pending_reqs or write requests in flight,
          * this segment must be finalized here in case we don't receive
          * AppendEntries RPCs anymore (could happen during a Snapshot install,
-         * causing the BarrierCb to never fire), but check that the callbacks that
-         * fired after completion of this write didn't already close the segment. */
+         * causing the BarrierCb to never fire), but check that the callbacks
+         * that fired after completion of this write didn't already close the
+         * segment. */
         uvAliveSegmentFinalize(s);
     }
 }
@@ -329,9 +357,13 @@ start:
         return 0;
     }
 
-    /* If there's a blocking barrier in progress, and it's not waiting for this segment
-     * to be finalized, let's wait. */
-    if (uv->barrier != NULL && segment->barrier != uv->barrier && uv->barrier->blocking) {
+    /* If there's a blocking barrier in progress, and it's not waiting for this
+     * segment to be finalized, let's wait.
+     *
+     * FIXME shouldn't we wait even if segment->barrier == uv->barrier, if there
+     * are other open segments associated with the same barrier? */
+    if (uv->barrier != NULL && segment->barrier != uv->barrier &&
+        uv->barrier->blocking) {
         return 0;
     }
 
@@ -413,6 +445,7 @@ static int uvAliveSegmentReady(struct uv *uv,
         ErrMsgWrapf(uv->io->errmsg, "setup writer for open-%llu", counter);
         return rv;
     }
+    UvWriterSetTracer(&segment->writer, uv->tracer);
     segment->counter = counter;
     return 0;
 }
@@ -436,6 +469,7 @@ static void uvAliveSegmentPrepareCb(struct uvPrepare *req, int status)
     }
 
     if (status != 0) {
+        tracef("prepare segment failed (%d)", status);
         rv = status;
         goto err;
     }
@@ -449,11 +483,13 @@ static void uvAliveSegmentPrepareCb(struct uvPrepare *req, int status)
 
     rv = uvAliveSegmentReady(uv, req->fd, req->counter, segment);
     if (rv != 0) {
+        tracef("prepare segment ready failed (%d)", rv);
         goto err;
     }
 
     rv = uvAppendMaybeStart(uv);
     if (rv != 0) {
+        tracef("prepare segment start failed (%d)", rv);
         goto err;
     }
 
@@ -583,6 +619,7 @@ static int uvAppendEnqueueRequest(struct uv *uv, struct uvAppend *append)
     assert(append->entries != NULL);
     assert(append->n > 0);
     assert(uv->append_next_index > 0);
+    tracef("enqueue %u entries", append->n);
 
     size = uvAppendSize(append);
 
@@ -614,6 +651,7 @@ static int uvAppendEnqueueRequest(struct uv *uv, struct uvAppend *append)
     append->segment = segment;
     QUEUE_PUSH(&uv->append_pending_reqs, &append->queue);
     uv->append_next_index += append->n;
+    tracef("set uv->append_next_index %llu", uv->append_next_index);
 
     return 0;
 
@@ -622,6 +660,25 @@ err:
     return rv;
 }
 
+/* Check that all entry buffers are 8-byte aligned */
+static int uvCheckEntryBuffersAligned(struct uv *uv,
+                                      const struct raft_entry entries[],
+                                      unsigned n)
+{
+    unsigned i;
+
+    for (i = 0; i < n; i++) {
+        if (entries[i].buf.len % 8) {
+            ErrMsgPrintf(uv->io->errmsg,
+                         "entry buffers must be 8-byte aligned");
+            Tracef(uv->tracer, "%s", uv->io->errmsg);
+            return RAFT_INVALID;
+        }
+    }
+
+    return 0;
+}
+
 int UvAppend(struct raft_io *io,
              struct raft_io_append *req,
              const struct raft_entry entries[],
@@ -645,6 +702,11 @@ int UvAppend(struct raft_io *io,
     append->n = n;
     req->cb = cb;
 
+    rv = uvCheckEntryBuffersAligned(uv, entries, n);
+    if (rv != 0) {
+        goto err_after_req_alloc;
+    }
+
     rv = uvAppendEnqueueRequest(uv, append);
     if (rv != 0) {
         goto err_after_req_alloc;
@@ -685,8 +747,7 @@ static void uvFinalizeCurrentAliveSegmentOnceIdle(struct uv *uv)
     /* Check if there are pending append requests targeted to the current
      * segment. */
     has_pending_reqs = false;
-    QUEUE_FOREACH(head, &uv->append_pending_reqs)
-    {
+    QUEUE_FOREACH (head, &uv->append_pending_reqs) {
         struct uvAppend *r = QUEUE_DATA(head, struct uvAppend, queue);
         if (r->segment == s) {
             has_pending_reqs = true;
@@ -715,40 +776,92 @@ bool UvBarrierReady(struct uv *uv)
     }
 
     queue *head;
-    QUEUE_FOREACH(head, &uv->append_segments)
-    {
+    QUEUE_FOREACH (head, &uv->append_segments) {
         struct uvAliveSegment *segment;
         segment = QUEUE_DATA(head, struct uvAliveSegment, queue);
         if (segment->barrier == uv->barrier) {
-                return false;
+            return false;
         }
     }
     return true;
 }
 
-int UvBarrier(struct uv *uv,
-              raft_index next_index,
-              struct UvBarrier *barrier,
-              UvBarrierCb cb)
+bool UvBarrierMaybeTrigger(struct UvBarrier *barrier)
 {
+    if (!barrier) {
+        return false;
+    }
+
+    if (!QUEUE_IS_EMPTY(&barrier->reqs)) {
+        queue *head;
+        struct UvBarrierReq *r;
+        head = QUEUE_HEAD(&barrier->reqs);
+        QUEUE_REMOVE(head);
+        r = QUEUE_DATA(head, struct UvBarrierReq, queue);
+        r->cb(r);
+        return true;
+    }
+
+    return false;
+}
+
+/* Used during cleanup. */
+static void uvBarrierTriggerAll(struct UvBarrier *barrier)
+{
+    while (UvBarrierMaybeTrigger(barrier)) {
+        ;
+    }
+}
+
+static struct UvBarrier *uvBarrierCreate(void)
+{
+    struct UvBarrier *barrier;
+    barrier = RaftHeapCalloc(1, sizeof(*barrier));
+    if (!barrier) {
+        return NULL;
+    }
+    barrier->blocking = false;
+    QUEUE_INIT(&barrier->reqs);
+    return barrier;
+}
+
+int UvBarrier(struct uv *uv, raft_index next_index, struct UvBarrierReq *req)
+{
+    /* The barrier to attach to. */
+    struct UvBarrier *barrier = NULL;
+    struct uvAliveSegment *segment = NULL;
     queue *head;
 
     assert(!uv->closing);
 
     /* The next entry will be appended at this index. */
     uv->append_next_index = next_index;
+    tracef("UvBarrier uv->append_next_index:%llu", uv->append_next_index);
 
     /* Arrange for all open segments not already involved in other barriers to
      * be finalized as soon as their append requests get completed and mark them
-     * as involved in this specific barrier request.  */
-    QUEUE_FOREACH(head, &uv->append_segments)
-    {
-        struct uvAliveSegment *segment;
+     * as involved in this specific barrier request. */
+    QUEUE_FOREACH (head, &uv->append_segments) {
         segment = QUEUE_DATA(head, struct uvAliveSegment, queue);
         if (segment->barrier != NULL) {
+            /* If a non-blocking barrier precedes this blocking request, we want
+             * to also block all future writes. */
+            if (req->blocking) {
+                segment->barrier->blocking = true;
+            }
             continue;
         }
+
+        if (!barrier) {
+            barrier = uvBarrierCreate();
+            if (!barrier) {
+                return RAFT_NOMEM;
+            }
+            /* And add the request to the barrier. */
+            UvBarrierAddReq(barrier, req);
+        }
         segment->barrier = barrier;
+
         if (segment == uvGetCurrentAliveSegment(uv)) {
             uvFinalizeCurrentAliveSegmentOnceIdle(uv);
             continue;
@@ -756,8 +869,32 @@ int UvBarrier(struct uv *uv,
         segment->finalize = true;
     }
 
-    barrier->cb = cb;
+    /* Unable to attach to a segment, because all segments are involved in a
+     * barrier, or there are no segments. */
+    if (barrier == NULL) {
+        /* Attach req to last segment barrier. */
+        if (segment != NULL) {
+            barrier = segment->barrier;
+            /* There is no segment, attach to uv->barrier. */
+        } else if (uv->barrier != NULL) {
+            barrier = uv->barrier;
+            /* There is no uv->barrier, make new one. */
+        } else {
+            barrier = uvBarrierCreate();
+            if (!barrier) {
+                return RAFT_NOMEM;
+            }
+        }
+        UvBarrierAddReq(barrier, req);
+    }
 
+    /* Let's not continue writing new entries if something down the line
+     * asked us to stop writing. */
+    if (uv->barrier != NULL && req->blocking) {
+        uv->barrier->blocking = true;
+    }
+
+    assert(barrier != NULL);
     if (uv->barrier == NULL) {
         uv->barrier = barrier;
         /* If there's no pending append-related activity, we can fire the
@@ -767,7 +904,8 @@ int UvBarrier(struct uv *uv,
         if (QUEUE_IS_EMPTY(&uv->append_segments) &&
             QUEUE_IS_EMPTY(&uv->finalize_reqs) &&
             uv->finalize_work.data == NULL) {
-            barrier->cb(barrier);
+            /* Not interested in return value. */
+            UvBarrierMaybeTrigger(barrier);
         }
     }
 
@@ -776,6 +914,16 @@ int UvBarrier(struct uv *uv,
 
 void UvUnblock(struct uv *uv)
 {
+    /* First fire all pending barrier requests. Unblock will be called again
+     * when that request's callback is fired.  */
+    if (UvBarrierMaybeTrigger(uv->barrier)) {
+        tracef("UvUnblock triggered barrier request callback.");
+        return;
+    }
+
+    /* All requests in barrier are finished. */
+    tracef("UvUnblock queue empty");
+    RaftHeapFree(uv->barrier);
     uv->barrier = NULL;
     if (uv->closing) {
         uvMaybeFireCloseCb(uv);
@@ -790,37 +938,79 @@ void UvUnblock(struct uv *uv)
     }
 }
 
+void UvBarrierAddReq(struct UvBarrier *barrier, struct UvBarrierReq *req)
+{
+    assert(barrier != NULL);
+    assert(req != NULL);
+    /* Once there's a blocking req, this barrier becomes blocking. */
+    barrier->blocking |= req->blocking;
+    QUEUE_PUSH(&barrier->reqs, &req->queue);
+}
+
 /* Fire all pending barrier requests, the barrier callback will notice that
  * we're closing and abort there. */
 static void uvBarrierClose(struct uv *uv)
 {
+    tracef("uv barrier close");
     struct UvBarrier *barrier = NULL;
     queue *head;
     assert(uv->closing);
-    QUEUE_FOREACH(head, &uv->append_segments)
-    {
+    QUEUE_FOREACH (head, &uv->append_segments) {
         struct uvAliveSegment *segment;
         segment = QUEUE_DATA(head, struct uvAliveSegment, queue);
-        if (segment->barrier != NULL && segment->barrier != barrier) {
+        if (segment->barrier != NULL && segment->barrier != barrier &&
+            segment->barrier != uv->barrier) {
             barrier = segment->barrier;
-            barrier->cb(barrier);
-            if (segment->barrier == uv->barrier) {
-                uv->barrier = NULL;
-            }
+            /* Fire all barrier cb's, this is safe because the barrier cb exits
+             * early when uv->closing is true. */
+            uvBarrierTriggerAll(barrier);
+            RaftHeapFree(barrier);
         }
+        /* The segment->barrier field is used:
+         *
+         * - by UvBarrierReady, to check whether it's time to invoke the barrier
+         *   callback after successfully finalizing a segment
+         * - by uvAppendMaybeStart, to see whether we should go ahead with
+         * writing to a segment even though a barrier is active because the
+         * barrier is waiting on that same segment to be finalized (but see the
+         * FIXME in that function)
+         * - to save a barrier for later, if UvBarrier was called when
+         * uv->barrier was already set
+         *
+         * If we're cancelling the barrier, we don't need to save it for later;
+         * the callback will not be invoked a second time in any case; and
+         * uvAppendMaybeStart won't be called while closing. So it's fine to
+         * clear segment->barrier here. */
         segment->barrier = NULL;
     }
 
-    /* There might still still be a current barrier set on uv->barrier, meaning
+    /* There might still be a current barrier set on uv->barrier, meaning
      * that the open segment it was associated with has started to be finalized
-     * and is not anymore in the append_segments queue. Let's cancel that
-     * too. */
+     * and is not anymore in the append_segments queue. Let's cancel all
+     * untriggered barrier request callbacks too. */
     if (uv->barrier != NULL) {
-        uv->barrier->cb(uv->barrier);
-        uv->barrier = NULL;
+        uvBarrierTriggerAll(uv->barrier);
+        /* Clear uv->barrier if there's no active work on the thread pool. When
+         * the work on the threadpool finishes, UvUnblock will notice
+         * we're closing, clear and free uv->barrier and call
+         * uvMaybeFireCloseCb. UnUnblock will not try to fire anymore barrier
+         * request callbacks because they were triggered in the line above. */
+        if (uv->snapshot_put_work.data == NULL &&
+            uv->truncate_work.data == NULL) {
+            RaftHeapFree(uv->barrier);
+            uv->barrier = NULL;
+        }
     }
 }
 
+static void uvAppendRetryCloseCb(uv_handle_t *handle)
+{
+    struct uv *uv = handle->data;
+    assert(uv->closing);
+    uv->append_retry.data = NULL;
+    uvMaybeFireCloseCb(uv);
+}
+
 void uvAppendClose(struct uv *uv)
 {
     struct uvAliveSegment *segment;
@@ -833,6 +1023,19 @@ void uvAppendClose(struct uv *uv)
 
     uvFinalizeCurrentAliveSegmentOnceIdle(uv);
 
+    if (uv->append_retry.data != NULL) {
+        /* If the timer data is not uv, then it's set to an alive segment, and
+         * we're currently waiting to retry a failed disk write. In that case,
+         * cancel all inflight requests. */
+        if (uv->append_retry.data != uv) {
+            uvAppendFinishWritingRequests(uv, RAFT_CANCELED);
+            uvAliveSegmentFinalize(uv->append_retry.data);
+            uv->append_retry.data = uv;
+        }
+        uv_timer_stop(&uv->append_retry);
+        uv_close((uv_handle_t *)&uv->append_retry, uvAppendRetryCloseCb);
+    }
+
     /* Also finalize the segments that we didn't write at all and are just
      * sitting in the append_segments queue waiting for writes against the
      * current segment to complete. */
@@ -846,3 +1049,29 @@ void uvAppendClose(struct uv *uv)
         uvAliveSegmentFinalize(segment);
     }
 }
+
+size_t UvAppendCapacity(struct uv *uv)
+{
+    size_t capacity = 0;
+    queue *head;
+
+    QUEUE_FOREACH (head, &uv->append_segments) {
+        struct uvAliveSegment *segment;
+        size_t size;
+        segment = QUEUE_DATA(head, struct uvAliveSegment, queue);
+
+        /* Skip segments that were not assigned a prepared segment yet. */
+        if (segment->counter == 0) {
+            continue;
+        }
+
+        size = segment->size;
+        if (size > uv->segment_size) {
+            size = uv->segment_size;
+        }
+
+        capacity += uv->segment_size - size;
+    }
+
+    return capacity;
+}
