@@ -1,5 +1,6 @@
 #include "uv_encoding.h"
 
+#include <limits.h>
 #include <string.h>
 
 #include "../include/raft/uv.h"
@@ -35,39 +36,50 @@ static size_t sizeofRequestVoteResultV1(void)
 
 static size_t sizeofRequestVoteResult(void)
 {
-    return sizeofRequestVoteResultV1() +
-           sizeof(uint64_t) /* Flags. */;
+    return sizeofRequestVoteResultV1() + /* Size of older version 1 message */
+           sizeof(uint8_t) +             /* Flags */
+           sizeof(uint8_t) +             /* Unused */
+           sizeof(uint16_t) +            /* Features */
+           sizeof(uint16_t) +            /* Capacity */
+           sizeof(uint16_t);             /* Unused */
 }
 
 static size_t sizeofAppendEntries(const struct raft_append_entries *p)
 {
-    return sizeof(uint64_t) + /* Leader's term. */
-           sizeof(uint64_t) + /* Leader ID */
-           sizeof(uint64_t) + /* Previous log entry index */
-           sizeof(uint64_t) + /* Previous log entry term */
-           sizeof(uint64_t) + /* Leader's commit index */
-           sizeof(uint64_t) + /* Number of entries in the batch */
-           16 * p->n_entries /* One header per entry */;
+    return sizeof(uint64_t) +                  /* Leader's term. */
+           sizeof(uint64_t) +                  /* Previous log entry index */
+           sizeof(uint64_t) +                  /* Previous log entry term */
+           sizeof(uint64_t) +                  /* Leader's commit index */
+           uvSizeofBatchHeader(p->n_entries) + /* Batch header */
+           sizeof(uint64_t);                   /* XXX: currently unused */
+}
+
+static size_t sizeofAppendEntriesResultV0(void)
+{
+    return sizeof(uint64_t) + /* Term. */
+           sizeof(uint64_t) + /* Success. */
+           sizeof(uint64_t);  /* Last log index. */
 }
 
 static size_t sizeofAppendEntriesResult(void)
 {
-    return sizeof(uint64_t) + /* Term. */
-           sizeof(uint64_t) + /* Success. */
-           sizeof(uint64_t) /* Last log index. */;
+    return sizeofAppendEntriesResultV0() + /* Size of older version 0 message */
+           sizeof(uint16_t) +              /* Server features. */
+           sizeof(uint16_t) +              /* Capacity. */
+           sizeof(uint32_t);               /* Unused */
 }
 
 static size_t sizeofInstallSnapshot(const struct raft_install_snapshot *p)
 {
     size_t conf_size = configurationEncodedSize(&p->conf);
     return sizeof(uint64_t) + /* Leader's term. */
-           sizeof(uint64_t) + /* Leader ID */
            sizeof(uint64_t) + /* Snapshot's last index */
            sizeof(uint64_t) + /* Term of last index */
            sizeof(uint64_t) + /* Configuration's index */
            sizeof(uint64_t) + /* Length of configuration */
            conf_size +        /* Configuration data */
-           sizeof(uint64_t);  /* Length of snapshot data */
+           sizeof(uint64_t) + /* Length of snapshot data */
+           sizeof(uint64_t);  /* XXX: currently unused */
 }
 
 static size_t sizeofTimeoutNow(void)
@@ -85,7 +97,7 @@ size_t uvSizeofBatchHeader(size_t n)
 
 static void encodeRequestVote(const struct raft_request_vote *p, void *buf)
 {
-    void *cursor = buf;
+    uint8_t *cursor = buf;
     uint64_t flags = 0;
 
     if (p->disrupt_leader) {
@@ -105,64 +117,80 @@ static void encodeRequestVote(const struct raft_request_vote *p, void *buf)
 static void encodeRequestVoteResult(const struct raft_request_vote_result *p,
                                     void *buf)
 {
-    void *cursor = buf;
-    uint64_t flags = 0;
+    uint8_t *cursor = buf;
+    uint8_t flags = 0;
 
-    if (p->pre_vote == raft_tribool_true) {
+    if (p->pre_vote) {
         flags |= (1 << 0);
     }
 
     bytePut64(&cursor, p->term);
     bytePut64(&cursor, p->vote_granted);
-    bytePut64(&cursor, flags);
+
+    bytePut8(&cursor, flags);
+    bytePut8(&cursor, 0);
+    bytePut16(&cursor, p->features);
+
+    bytePut16(&cursor, p->capacity);
+    bytePut16(&cursor, 0);
 }
 
 static void encodeAppendEntries(const struct raft_append_entries *p, void *buf)
 {
-    void *cursor;
+    uint8_t *cursor;
 
     cursor = buf;
 
     bytePut64(&cursor, p->term);           /* Leader's term. */
-    bytePut64(&cursor, p->prev_log_index); /* Previous index. */
-    bytePut64(&cursor, p->prev_log_term);  /* Previous term. */
-    bytePut64(&cursor, p->leader_commit);  /* Commit index. */
+    bytePut64(&cursor, p->prev_log_index); /* Previous log entry index. */
+    bytePut64(&cursor, p->prev_log_term);  /* Previous log entry term. */
+    bytePut64(&cursor, p->leader_commit);  /* Leader's commit index. */
 
-    uvEncodeBatchHeader(p->entries, p->n_entries, cursor);
+    uvEncodeBatchHeader(p->entries, p->n_entries, cursor); /* Batch header */
+
+    cursor = (uint8_t *)cursor + uvSizeofBatchHeader(p->n_entries);
+    bytePut64(&cursor, 0); /* XXX: currently unused */
 }
 
 static void encodeAppendEntriesResult(
     const struct raft_append_entries_result *p,
     void *buf)
 {
-    void *cursor = buf;
+    uint8_t *cursor = buf;
 
     bytePut64(&cursor, p->term);
     bytePut64(&cursor, p->rejected);
     bytePut64(&cursor, p->last_log_index);
+    bytePut16(&cursor, p->features);
+    bytePut16(&cursor, p->capacity);
+    bytePut32(&cursor, 0 /* Unused */);
 }
 
 static void encodeInstallSnapshot(const struct raft_install_snapshot *p,
                                   void *buf)
 {
-    void *cursor;
+    uint8_t *cursor;
     size_t conf_size = configurationEncodedSize(&p->conf);
 
     cursor = buf;
 
-    bytePut64(&cursor, p->term);       /* Leader's term. */
-    bytePut64(&cursor, p->last_index); /* Snapshot last index. */
-    bytePut64(&cursor, p->last_term);  /* Term of last index. */
-    bytePut64(&cursor, p->conf_index); /* Configuration index. */
-    bytePut64(&cursor, conf_size);     /* Configuration length. */
-    configurationEncodeToBuf(&p->conf, cursor);
+    bytePut64(&cursor, p->term);       /* Leader's term */
+    bytePut64(&cursor, p->last_index); /* Snapshot's last index */
+    bytePut64(&cursor, p->last_term);  /* Term of last index */
+    bytePut64(&cursor, p->conf_index); /* Configuration's index */
+    bytePut64(&cursor, conf_size);     /* Length of configuration */
+
+    configurationEncodeToBuf(&p->conf, cursor); /* Configuration data */
     cursor = (uint8_t *)cursor + conf_size;
-    bytePut64(&cursor, p->data.len); /* Snapshot data size. */
+
+    bytePut64(&cursor, p->data.len); /* Length of snapshot data */
+
+    bytePut64(&cursor, 0); /* XXX: currently unused */
 }
 
 static void encodeTimeoutNow(const struct raft_timeout_now *p, void *buf)
 {
-    void *cursor = buf;
+    uint8_t *cursor = buf;
 
     bytePut64(&cursor, p->term);
     bytePut64(&cursor, p->last_log_index);
@@ -174,29 +202,36 @@ int uvEncodeMessage(const struct raft_message *message,
                     unsigned *n_bufs)
 {
     uv_buf_t header;
-    void *cursor;
+    uint8_t *cursor;
+    int version;
 
     /* Figure out the length of the header for this request and allocate a
      * buffer for it. */
     header.len = RAFT_IO_UV__PREAMBLE_SIZE;
     switch (message->type) {
-        case RAFT_IO_REQUEST_VOTE:
+        case RAFT_REQUEST_VOTE:
             header.len += sizeofRequestVote();
+            version = message->request_vote.version;
             break;
-        case RAFT_IO_REQUEST_VOTE_RESULT:
+        case RAFT_REQUEST_VOTE_RESULT:
             header.len += sizeofRequestVoteResult();
+            version = message->request_vote_result.version;
             break;
-        case RAFT_IO_APPEND_ENTRIES:
+        case RAFT_APPEND_ENTRIES:
             header.len += sizeofAppendEntries(&message->append_entries);
+            version = message->append_entries.version;
             break;
-        case RAFT_IO_APPEND_ENTRIES_RESULT:
+        case RAFT_APPEND_ENTRIES_RESULT:
             header.len += sizeofAppendEntriesResult();
+            version = message->append_entries_result.version;
             break;
-        case RAFT_IO_INSTALL_SNAPSHOT:
+        case RAFT_INSTALL_SNAPSHOT:
             header.len += sizeofInstallSnapshot(&message->install_snapshot);
+            version = message->install_snapshot.version;
             break;
-        case RAFT_IO_TIMEOUT_NOW:
+        case RAFT_TIMEOUT_NOW:
             header.len += sizeofTimeoutNow();
+            version = message->timeout_now.version;
             break;
         default:
             return RAFT_MALFORMED;
@@ -207,30 +242,35 @@ int uvEncodeMessage(const struct raft_message *message,
         goto oom;
     }
 
-    cursor = header.base;
+    cursor = (uint8_t *)header.base;
+
+    /* Encode the request preamble, with message type, version and size. */
+    bytePut8(&cursor, (uint8_t)message->type);
+    bytePut8(&cursor, 0);
+    bytePut8(&cursor, (uint8_t)version);
+    bytePut8(&cursor, 0);
+    bytePut32(&cursor, 0);
 
-    /* Encode the request preamble, with message type and message size. */
-    bytePut64(&cursor, message->type);
     bytePut64(&cursor, header.len - RAFT_IO_UV__PREAMBLE_SIZE);
 
     /* Encode the request header. */
     switch (message->type) {
-        case RAFT_IO_REQUEST_VOTE:
+        case RAFT_REQUEST_VOTE:
             encodeRequestVote(&message->request_vote, cursor);
             break;
-        case RAFT_IO_REQUEST_VOTE_RESULT:
+        case RAFT_REQUEST_VOTE_RESULT:
             encodeRequestVoteResult(&message->request_vote_result, cursor);
             break;
-        case RAFT_IO_APPEND_ENTRIES:
+        case RAFT_APPEND_ENTRIES:
             encodeAppendEntries(&message->append_entries, cursor);
             break;
-        case RAFT_IO_APPEND_ENTRIES_RESULT:
+        case RAFT_APPEND_ENTRIES_RESULT:
             encodeAppendEntriesResult(&message->append_entries_result, cursor);
             break;
-        case RAFT_IO_INSTALL_SNAPSHOT:
+        case RAFT_INSTALL_SNAPSHOT:
             encodeInstallSnapshot(&message->install_snapshot, cursor);
             break;
-        case RAFT_IO_TIMEOUT_NOW:
+        case RAFT_TIMEOUT_NOW:
             encodeTimeoutNow(&message->timeout_now, cursor);
             break;
     };
@@ -238,12 +278,12 @@ int uvEncodeMessage(const struct raft_message *message,
     *n_bufs = 1;
 
     /* For AppendEntries request we also send the entries payload. */
-    if (message->type == RAFT_IO_APPEND_ENTRIES) {
+    if (message->type == RAFT_APPEND_ENTRIES) {
         *n_bufs += message->append_entries.n_entries;
     }
 
     /* For InstallSnapshot request we also send the snapshot payload. */
-    if (message->type == RAFT_IO_INSTALL_SNAPSHOT) {
+    if (message->type == RAFT_INSTALL_SNAPSHOT) {
         *n_bufs += 1;
     }
 
@@ -254,7 +294,7 @@ int uvEncodeMessage(const struct raft_message *message,
 
     (*bufs)[0] = header;
 
-    if (message->type == RAFT_IO_APPEND_ENTRIES) {
+    if (message->type == RAFT_APPEND_ENTRIES) {
         unsigned i;
         for (i = 0; i < message->append_entries.n_entries; i++) {
             const struct raft_entry *entry =
@@ -264,7 +304,7 @@ int uvEncodeMessage(const struct raft_message *message,
         }
     }
 
-    if (message->type == RAFT_IO_INSTALL_SNAPSHOT) {
+    if (message->type == RAFT_INSTALL_SNAPSHOT) {
         (*bufs)[1].base = message->install_snapshot.data.base;
         (*bufs)[1].len = message->install_snapshot.data.len;
     }
@@ -283,7 +323,7 @@ void uvEncodeBatchHeader(const struct raft_entry *entries,
                          void *buf)
 {
     unsigned i;
-    void *cursor = buf;
+    uint8_t *cursor = buf;
 
     /* Number of entries in the batch, little endian */
     bytePut64(&cursor, n);
@@ -304,44 +344,73 @@ void uvEncodeBatchHeader(const struct raft_entry *entries,
     }
 }
 
-static void decodeRequestVote(const uv_buf_t *buf, struct raft_request_vote *p)
+static void decodeRequestVote(int version,
+                              const uv_buf_t *buf,
+                              struct raft_request_vote *p)
 {
-    const void *cursor;
+    const uint8_t *cursor;
 
-    cursor = buf->base;
+    cursor = (void *)buf->base;
 
+    /* If the version is 0 the message was sent by a server that
+     * does not encodes the version byte in the preamble. */
+    if (version == 0) {
+        if (buf->len == sizeofRequestVoteV1()) {
+            version = 1;
+        } else {
+            version = 2;
+        }
+    }
+
+    p->version = version;
     p->term = byteGet64(&cursor);
     p->candidate_id = byteGet64(&cursor);
     p->last_log_index = byteGet64(&cursor);
     p->last_log_term = byteGet64(&cursor);
+    p->disrupt_leader = false;
+    p->pre_vote = false;
 
-    /* Support for legacy request vote that doesn't have disrupt_leader. */
-    if (buf->len == sizeofRequestVoteV1()) {
-        p->disrupt_leader = false;
-        p->pre_vote = false;
-    } else {
+    if (p->version >= 1) {
         uint64_t flags = byteGet64(&cursor);
         p->disrupt_leader = (bool)(flags & 1 << 0);
         p->pre_vote = (bool)(flags & 1 << 1);
     }
 }
 
-static void decodeRequestVoteResult(const uv_buf_t *buf,
+static void decodeRequestVoteResult(int version,
+                                    const uv_buf_t *buf,
                                     struct raft_request_vote_result *p)
 {
-    const void *cursor;
+    const uint8_t *cursor;
 
-    cursor = buf->base;
+    cursor = (void *)buf->base;
 
+    /* If the version is 0 the message was sent by a server that
+     * does not encodes the version byte in the preamble. */
+    if (version == 0) {
+        if (buf->len == sizeofRequestVoteResultV1()) {
+            version = 1;
+        } else {
+            version = 2;
+        }
+    }
+
+    p->version = version;
     p->term = byteGet64(&cursor);
     p->vote_granted = byteGet64(&cursor);
+    p->features = 0;
+    p->capacity = 0;
 
-    /* Support legacy RequestVoteResultV1 */
-    p->pre_vote = raft_tribool_unknown;
-
-    if (buf->len > sizeofRequestVoteResultV1()) {
-        uint64_t flags = byteGet64(&cursor);
-        p->pre_vote = TO_RAFT_TRIBOOL(flags & (1 << 0));
+    if (p->version >= 2) {
+        uint8_t flags = byteGet8(&cursor);
+        uint8_t unused = byteGet8(&cursor);
+        uint16_t features = byteGet16(&cursor);
+        uint16_t capacity = byteGet16(&cursor);
+        p->version = 2;
+        p->pre_vote = (flags & (1 << 0));
+        (void)unused;
+        p->features = features;
+        p->capacity = capacity;
     }
 }
 
@@ -349,7 +418,7 @@ int uvDecodeBatchHeader(const void *batch,
                         struct raft_entry **entries,
                         unsigned *n)
 {
-    const void *cursor = batch;
+    const uint8_t *cursor = batch;
     size_t i;
     int rv;
 
@@ -397,17 +466,19 @@ err:
     return rv;
 }
 
-static int decodeAppendEntries(const uv_buf_t *buf,
+static int decodeAppendEntries(int version,
+                               const uv_buf_t *buf,
                                struct raft_append_entries *args)
 {
-    const void *cursor;
+    const uint8_t *cursor;
     int rv;
 
     assert(buf != NULL);
     assert(args != NULL);
 
-    cursor = buf->base;
+    cursor = (void *)buf->base;
 
+    args->version = version;
     args->term = byteGet64(&cursor);
     args->prev_log_index = byteGet64(&cursor);
     args->prev_log_term = byteGet64(&cursor);
@@ -421,37 +492,57 @@ static int decodeAppendEntries(const uv_buf_t *buf,
     return 0;
 }
 
-static void decodeAppendEntriesResult(const uv_buf_t *buf,
+static void decodeAppendEntriesResult(int version,
+                                      const uv_buf_t *buf,
                                       struct raft_append_entries_result *p)
 {
-    const void *cursor;
+    const uint8_t *cursor;
 
-    cursor = buf->base;
+    cursor = (void *)buf->base;
 
+    /* If the version is 0 the message was sent by a server that
+     * does not encodes the version byte in the preamble. */
+    if (version == 0) {
+        if (buf->len > sizeofAppendEntriesResultV0()) {
+            version = 2;
+        }
+    }
+
+    p->version = version;
     p->term = byteGet64(&cursor);
     p->rejected = byteGet64(&cursor);
     p->last_log_index = byteGet64(&cursor);
+    p->features = 0;
+    p->capacity = 0;
+    if (p->version >= 1) {
+        p->features = byteGet16(&cursor);
+    }
+    if (p->version >= 2) {
+        p->capacity = byteGet16(&cursor);
+    }
 }
 
-static int decodeInstallSnapshot(const uv_buf_t *buf,
+static int decodeInstallSnapshot(int version,
+                                 const uv_buf_t *buf,
                                  struct raft_install_snapshot *args)
 {
-    const void *cursor;
+    const uint8_t *cursor;
     struct raft_buffer conf;
     int rv;
 
     assert(buf != NULL);
     assert(args != NULL);
 
-    cursor = buf->base;
+    cursor = (void *)buf->base;
 
+    args->version = version;
     args->term = byteGet64(&cursor);
     args->last_index = byteGet64(&cursor);
     args->last_term = byteGet64(&cursor);
     args->conf_index = byteGet64(&cursor);
     conf.len = (size_t)byteGet64(&cursor);
     conf.base = (void *)cursor;
-    configurationInit(&args->conf);
+
     rv = configurationDecode(&conf, &args->conf);
     if (rv != 0) {
         return rv;
@@ -464,16 +555,18 @@ static int decodeInstallSnapshot(const uv_buf_t *buf,
 
 static void decodeTimeoutNow(const uv_buf_t *buf, struct raft_timeout_now *p)
 {
-    const void *cursor;
+    const uint8_t *cursor;
 
-    cursor = buf->base;
+    cursor = (void *)buf->base;
 
+    p->version = 0;
     p->term = byteGet64(&cursor);
     p->last_log_index = byteGet64(&cursor);
     p->last_log_term = byteGet64(&cursor);
 }
 
-int uvDecodeMessage(const unsigned long type,
+int uvDecodeMessage(uint8_t type,
+                    uint8_t version,
                     const uv_buf_t *header,
                     struct raft_message *message,
                     size_t *payload_len)
@@ -481,33 +574,36 @@ int uvDecodeMessage(const unsigned long type,
     unsigned i;
     int rv = 0;
 
-    /* TODO: check type overflow */
+    memset(message, 0, sizeof(*message));
     message->type = (unsigned short)type;
 
     *payload_len = 0;
 
     /* Decode the header. */
     switch (type) {
-        case RAFT_IO_REQUEST_VOTE:
-            decodeRequestVote(header, &message->request_vote);
+        case RAFT_REQUEST_VOTE:
+            decodeRequestVote(version, header, &message->request_vote);
             break;
-        case RAFT_IO_REQUEST_VOTE_RESULT:
-            decodeRequestVoteResult(header, &message->request_vote_result);
+        case RAFT_REQUEST_VOTE_RESULT:
+            decodeRequestVoteResult(version, header,
+                                    &message->request_vote_result);
             break;
-        case RAFT_IO_APPEND_ENTRIES:
-            rv = decodeAppendEntries(header, &message->append_entries);
+        case RAFT_APPEND_ENTRIES:
+            rv = decodeAppendEntries(version, header, &message->append_entries);
             for (i = 0; i < message->append_entries.n_entries; i++) {
                 *payload_len += message->append_entries.entries[i].buf.len;
             }
             break;
-        case RAFT_IO_APPEND_ENTRIES_RESULT:
-            decodeAppendEntriesResult(header, &message->append_entries_result);
+        case RAFT_APPEND_ENTRIES_RESULT:
+            decodeAppendEntriesResult(version, header,
+                                      &message->append_entries_result);
             break;
-        case RAFT_IO_INSTALL_SNAPSHOT:
-            rv = decodeInstallSnapshot(header, &message->install_snapshot);
+        case RAFT_INSTALL_SNAPSHOT:
+            rv = decodeInstallSnapshot(version, header,
+                                       &message->install_snapshot);
             *payload_len += message->install_snapshot.data.len;
             break;
-        case RAFT_IO_TIMEOUT_NOW:
+        case RAFT_TIMEOUT_NOW:
             decodeTimeoutNow(header, &message->timeout_now);
             break;
         default:
@@ -548,42 +644,3 @@ void uvDecodeEntriesBatch(uint8_t *batch,
         }
     }
 }
-
-int uvEncodeSnapshotMeta(const struct raft_configuration *conf,
-                         raft_index conf_index,
-                         struct raft_buffer *buf)
-{
-    size_t conf_len;
-    void *cursor;
-    uint64_t *header;
-    void *conf_buf;
-    unsigned crc;
-
-    conf_len = configurationEncodedSize(conf);
-
-    buf->len = sizeof(*header) * 4; /* Format, CRC, configuration index/len */
-    buf->len += conf_len;
-    buf->base = raft_malloc(buf->len);
-    if (buf->base == NULL) {
-        return RAFT_NOMEM;
-    }
-
-    header = buf->base;
-    conf_buf = header + 4;
-
-    configurationEncodeToBuf(conf, conf_buf);
-
-    cursor = header;
-    bytePut64(&cursor, UV__DISK_FORMAT);
-    bytePut64(&cursor, 0);
-    bytePut64(&cursor, conf_index);
-    bytePut64(&cursor, conf_len);
-
-    crc = byteCrc32(&header[2], sizeof(uint64_t) * 2, 0); /* Conf index/len */
-    crc = byteCrc32(conf_buf, conf_len, crc);
-
-    cursor = &header[1];
-    bytePut64(&cursor, crc);
-
-    return 0;
-}
