@@ -6,7 +6,6 @@
 #include <unistd.h>
 
 #include "assert.h"
-#include "compress.h"
 #include "err.h"
 #include "heap.h"
 #include "uv_os.h"
@@ -78,7 +77,10 @@ int UvFsFileExists(const char *dir,
     char path[UV__PATH_SZ];
     int rv;
 
-    UvOsJoin(dir, filename, path);
+    rv = UvOsJoin(dir, filename, path);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
 
     rv = UvOsStat(path, &sb);
     if (rv != 0) {
@@ -106,7 +108,10 @@ int UvFsFileSize(const char *dir,
     char path[UV__PATH_SZ];
     int rv;
 
-    UvOsJoin(dir, filename, path);
+    rv = UvOsJoin(dir, filename, path);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
 
     rv = UvOsStat(path, &sb);
     if (rv != 0) {
@@ -144,7 +149,10 @@ static int uvFsOpenFile(const char *dir,
 {
     char path[UV__PATH_SZ];
     int rv;
-    UvOsJoin(dir, filename, path);
+    rv = UvOsJoin(dir, filename, path);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
     rv = UvOsOpen(path, flags, mode, fd);
     if (rv != 0) {
         UvOsErrMsg(errmsg, "open", rv);
@@ -160,34 +168,20 @@ int UvFsOpenFileForReading(const char *dir,
 {
     char path[UV__PATH_SZ];
     int flags = O_RDONLY;
+    int rv;
 
-    UvOsJoin(dir, filename, path);
+    rv = UvOsJoin(dir, filename, path);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
 
     return uvFsOpenFile(dir, filename, flags, 0, fd, errmsg);
 }
 
-int UvFsAllocateFile(const char *dir,
-                     const char *filename,
-                     size_t size,
-                     uv_file *fd,
-                     char *errmsg)
+static int uvFsAllocate(uv_file fd, size_t size, char *errmsg)
 {
-    char path[UV__PATH_SZ];
-    int flags = O_WRONLY | O_CREAT | O_EXCL; /* Common open flags */
-    int rv = 0;
-
-    UvOsJoin(dir, filename, path);
-
-    /* TODO: use RWF_DSYNC instead, if available. */
-    flags |= O_DSYNC;
-
-    rv = uvFsOpenFile(dir, filename, flags, S_IRUSR | S_IWUSR, fd, errmsg);
-    if (rv != 0) {
-        goto err;
-    }
-
-    /* Allocate the desired size. */
-    rv = UvOsFallocate(*fd, 0, (off_t)size);
+    int rv;
+    rv = UvOsFallocate(fd, 0, (off_t)size);
     if (rv != 0) {
         switch (rv) {
             case UV_ENOSPC:
@@ -200,6 +194,33 @@ int UvFsAllocateFile(const char *dir,
                 rv = RAFT_IOERR;
                 break;
         }
+    }
+    return rv;
+}
+
+int UvFsAllocateFile(const char *dir,
+                     const char *filename,
+                     size_t size,
+                     uv_file *fd,
+                     char *errmsg)
+{
+    char path[UV__PATH_SZ];
+    int flags = O_WRONLY | O_CREAT | O_EXCL; /* Common open flags */
+    int rv = 0;
+
+    rv = UvOsJoin(dir, filename, path);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
+
+    rv = uvFsOpenFile(dir, filename, flags, S_IRUSR | S_IWUSR, fd, errmsg);
+    if (rv != 0) {
+        goto err;
+    }
+
+    /* Allocate the desired size. */
+    rv = uvFsAllocate(*fd, size, errmsg);
+    if (rv != 0) {
         goto err_after_open;
     }
 
@@ -213,6 +234,97 @@ err:
     return rv;
 }
 
+int UvFsCreateTempFile(const char *dir,
+                       struct raft_buffer *bufs,
+                       unsigned n_bufs,
+                       uv_file *fd,
+                       char *errmsg)
+{
+    unsigned i;
+    size_t size = 0;
+    int rv;
+
+    for (i = 0; i < n_bufs; i++) {
+        size += bufs[i].len;
+    }
+
+    rv = uvFsOpenFile(dir, "", O_TMPFILE | O_WRONLY, S_IRUSR | S_IWUSR, fd,
+                      errmsg);
+    if (rv != 0) {
+        goto err;
+    }
+
+    rv = uvFsAllocate(*fd, size, errmsg);
+    if (rv != 0) {
+        goto err_after_open;
+    }
+
+    rv = UvOsWrite(*fd, (const uv_buf_t *)bufs, n_bufs, 0);
+    if (rv != (int)(size)) {
+        if (rv < 0) {
+            UvOsErrMsg(errmsg, "write", rv);
+        } else {
+            ErrMsgPrintf(errmsg, "short write: only %d bytes written", rv);
+        }
+        rv = RAFT_IOERR;
+        goto err_after_open;
+    }
+
+    rv = UvOsFsync(*fd);
+    if (rv != 0) {
+        UvOsErrMsg(errmsg, "fsync", rv);
+        rv = RAFT_IOERR;
+        goto err_after_open;
+    }
+
+    return 0;
+
+err_after_open:
+    UvOsClose(*fd);
+err:
+    assert(rv != 0);
+    return rv;
+}
+
+int UvFsFinalizeTempFile(uv_file fd,
+                         const char *dir,
+                         const char *filename,
+                         char *errmsg)
+{
+    char path[UV__PATH_SZ];
+    char procpath[PATH_MAX];
+    int rv;
+
+    rv = UvOsJoin(dir, filename, path);
+    if (rv != 0) {
+        rv = RAFT_INVALID;
+        goto err_before_close;
+    }
+
+    snprintf(procpath, PATH_MAX, "/proc/self/fd/%d", fd);
+    rv = UvOsLinkat(AT_FDCWD, procpath, AT_FDCWD, path, AT_SYMLINK_FOLLOW);
+    if (rv != 0) {
+        UvOsErrMsg(errmsg, "linkat", rv);
+        rv = RAFT_IOERR;
+        goto err_before_close;
+    }
+
+    rv = UvOsClose(fd);
+    if (rv != 0) {
+        UvOsErrMsg(errmsg, "close", rv);
+        rv = RAFT_IOERR;
+        goto err;
+    }
+
+    return 0;
+
+err_before_close:
+    UvOsClose(fd);
+err:
+    assert(rv != 0);
+    return rv;
+}
+
 static int uvFsWriteFile(const char *dir,
                          const char *filename,
                          int flags,
@@ -266,7 +378,7 @@ int UvFsMakeFile(const char *dir,
                  char *errmsg)
 {
     int rv;
-    char tmp_filename[UV__FILENAME_LEN+1] = {0};
+    char tmp_filename[UV__FILENAME_LEN + 1] = {0};
     char path[UV__PATH_SZ] = {0};
     char tmp_path[UV__PATH_SZ] = {0};
 
@@ -274,7 +386,7 @@ int UvFsMakeFile(const char *dir,
      * TODO as of libuv 1.34.0, use `uv_fs_mkstemp` */
     size_t sz = sizeof(tmp_filename);
     rv = snprintf(tmp_filename, sz, TMP_FILE_FMT, filename);
-    if (rv < 0 || rv >= (int) sz) {
+    if (rv < 0 || rv >= (int)sz) {
         return rv;
     }
     int flags = UV_FS_O_WRONLY | UV_FS_O_CREAT | UV_FS_O_EXCL;
@@ -297,8 +409,14 @@ int UvFsMakeFile(const char *dir,
     /* Rename the temp file. Remark that there is a race between the existence
      * check and the rename, there is no `renameat2` equivalent in libuv.
      * However, in the current implementation this should pose no problems.*/
-    UvOsJoin(dir, tmp_filename, tmp_path);
-    UvOsJoin(dir, filename, path);
+    rv = UvOsJoin(dir, tmp_filename, tmp_path);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
+    rv = UvOsJoin(dir, filename, path);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
     rv = UvOsRename(tmp_path, path);
     if (rv != 0) {
         UvOsErrMsg(errmsg, "rename", rv);
@@ -331,7 +449,10 @@ int UvFsMakeOrOverwriteFile(const char *dir,
     uv_file fd;
     int rv;
 
-    UvOsJoin(dir, filename, path);
+    rv = UvOsJoin(dir, filename, path);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
 
 open:
     rv = UvOsOpen(path, flags, mode, &fd);
@@ -392,16 +513,26 @@ err:
 
 int UvFsReadInto(uv_file fd, struct raft_buffer *buf, char *errmsg)
 {
-    int rv;
+    ssize_t rv;
+    size_t offset = 0;
+
     /* TODO: use uv_fs_read() */
-    rv = (int)read(fd, buf->base, buf->len);
-    if (rv == -1) {
-        UvOsErrMsg(errmsg, "read", -errno);
-        return RAFT_IOERR;
+    while (offset < buf->len) {
+        rv = read(fd, (char *)buf->base + offset, buf->len - offset);
+        if (rv == -1) {
+            UvOsErrMsg(errmsg, "read", -errno);
+            return RAFT_IOERR;
+        }
+        /* EOF. Don't think this is reachable, but just make very sure
+         * we don't loop forever. */
+        if (rv == 0) {
+            break;
+        }
+        assert(rv > 0);
+        offset += (size_t)rv;
     }
-    assert(rv >= 0);
-    if ((size_t)rv < buf->len) {
-        ErrMsgPrintf(errmsg, "short read: %d bytes instead of %zu", rv,
+    if (offset < buf->len) {
+        ErrMsgPrintf(errmsg, "short read: %zu bytes instead of %zu", offset,
                      buf->len);
         return RAFT_IOERR;
     }
@@ -418,7 +549,10 @@ int UvFsReadFile(const char *dir,
     uv_file fd;
     int rv;
 
-    UvOsJoin(dir, filename, path);
+    rv = UvOsJoin(dir, filename, path);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
 
     rv = UvOsStat(path, &sb);
     if (rv != 0) {
@@ -433,7 +567,7 @@ int UvFsReadFile(const char *dir,
     }
 
     buf->len = (size_t)sb.st_size;
-    buf->base = HeapMalloc(buf->len);
+    buf->base = RaftHeapMalloc(buf->len);
     if (buf->base == NULL) {
         ErrMsgOom(errmsg);
         rv = RAFT_NOMEM;
@@ -450,7 +584,7 @@ int UvFsReadFile(const char *dir,
     return 0;
 
 err_after_buf_alloc:
-    HeapFree(buf->base);
+    RaftHeapFree(buf->base);
 err_after_open:
     UvOsClose(fd);
 err:
@@ -466,7 +600,10 @@ int UvFsReadFileInto(const char *dir,
     uv_file fd;
     int rv;
 
-    UvOsJoin(dir, filename, path);
+    rv = UvOsJoin(dir, filename, path);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
 
     rv = uvFsOpenFile(dir, filename, O_RDONLY, 0, &fd, errmsg);
     if (rv != 0) {
@@ -492,7 +629,10 @@ int UvFsRemoveFile(const char *dir, const char *filename, char *errmsg)
 {
     char path[UV__PATH_SZ];
     int rv;
-    UvOsJoin(dir, filename, path);
+    rv = UvOsJoin(dir, filename, path);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
     rv = UvOsUnlink(path);
     if (rv != 0) {
         UvOsErrMsg(errmsg, "unlink", rv);
@@ -501,6 +641,33 @@ int UvFsRemoveFile(const char *dir, const char *filename, char *errmsg)
     return 0;
 }
 
+int UvFsRenameFile(const char *dir,
+                   const char *filename1,
+                   const char *filename2,
+                   char *errmsg)
+{
+    char path1[UV__PATH_SZ];
+    char path2[UV__PATH_SZ];
+    int rv;
+
+    rv = UvOsJoin(dir, filename1, path1);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
+    rv = UvOsJoin(dir, filename2, path2);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
+
+    rv = UvOsRename(path1, path2);
+    if (rv != 0) {
+        UvOsErrMsg(errmsg, "rename", rv);
+        return rv;
+    }
+
+    return 0;
+}
+
 int UvFsTruncateAndRenameFile(const char *dir,
                               size_t size,
                               const char *filename1,
@@ -512,8 +679,14 @@ int UvFsTruncateAndRenameFile(const char *dir,
     uv_file fd;
     int rv;
 
-    UvOsJoin(dir, filename1, path1);
-    UvOsJoin(dir, filename2, path2);
+    rv = UvOsJoin(dir, filename1, path1);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
+    rv = UvOsJoin(dir, filename2, path2);
+    if (rv != 0) {
+        return RAFT_INVALID;
+    }
 
     /* Truncate and rename. */
     rv = UvOsOpen(path1, UV_FS_O_RDWR, 0, &fd);
@@ -626,7 +799,6 @@ static int probeDirectIO(int fd, size_t *size, char *errmsg)
     return 0;
 }
 
-#if defined(RWF_NOWAIT)
 /* Check if fully non-blocking async I/O is possible on the given fd. */
 static int probeAsyncIO(int fd, size_t size, bool *ok, char *errmsg)
 {
@@ -657,7 +829,7 @@ static int probeAsyncIO(int fd, size_t size, bool *ok, char *errmsg)
     /* Prepare the KAIO request object */
     memset(&iocb, 0, sizeof iocb);
     iocb.aio_lio_opcode = IOCB_CMD_PWRITE;
-    *((void **)(&iocb.aio_buf)) = buf;
+    iocb.aio_buf = (uintptr_t)buf;
     iocb.aio_nbytes = size;
     iocb.aio_offset = 0;
     iocb.aio_fildes = (uint32_t)fd;
@@ -704,15 +876,23 @@ static int probeAsyncIO(int fd, size_t size, bool *ok, char *errmsg)
         assert(event.res == (int)size);
         *ok = true;
     } else {
-        /* UNTESTED: this should basically fail only because of disk errors,
-         * since we allocated the file with posix_fallocate and the block size
-         * is supposed to be correct. */
-        *ok = false;
+        if (event.res == -EAGAIN) {
+            /* UNTESTED: starting from around kernel version 6, xfs has started
+             * to occasionally fail with EAGAIN, presumeably because the write
+             * would block in some way. We still want to try submitting writes
+             * asynchronously in that case, and we'll deal with retries in the
+             * writer. */
+            *ok = true;
+        } else {
+            /* UNTESTED: this should basically fail only because of disk errors,
+             * since we allocated the file with posix_fallocate and the block
+             * size is supposed to be correct. */
+            *ok = false;
+        }
     }
 
     return 0;
 }
-#endif /* RWF_NOWAIT */
 
 #define UV__FS_PROBE_FILE ".probe"
 #define UV__FS_PROBE_FILE_SIZE 4096
@@ -739,14 +919,10 @@ int UvFsProbeCapabilities(const char *dir,
     /* Check if we can use direct I/O. */
     rv = probeDirectIO(fd, direct, errmsg);
     if (rv != 0) {
+        ErrMsgWrapf(errmsg, "probe Direct I/O");
         goto err_after_file_open;
     }
 
-#if !defined(RWF_NOWAIT)
-    /* We can't have fully async I/O, since io_submit might potentially block.
-     */
-    *async = false;
-#else
     /* If direct I/O is not possible, we can't perform fully asynchronous
      * I/O, because io_submit might potentially block. */
     if (*direct == 0) {
@@ -755,13 +931,11 @@ int UvFsProbeCapabilities(const char *dir,
     }
     rv = probeAsyncIO(fd, *direct, async, errmsg);
     if (rv != 0) {
+        ErrMsgWrapf(errmsg, "probe Async I/O");
         goto err_after_file_open;
     }
-#endif /* RWF_NOWAIT */
 
-#if defined(RWF_NOWAIT)
 out:
-#endif /* RWF_NOWAIT */
     close(fd);
     return 0;
 
