@@ -412,7 +412,8 @@ int uvSegmentLoadClosed(struct uv *uv,
         if (rv != 0) {
             ErrMsgWrapf(uv->io->errmsg, "entries batch %u starting at byte %zu",
                         i, offset);
-            goto err_after_read;
+            /* Clean up the last allocation from extendEntries. */
+            goto err_after_extend_entries;
         }
         rv = extendEntries(tmp_entries, tmp_n, entries, n);
         if (rv != 0) {
@@ -594,7 +595,8 @@ done:
 
         /* At least one entry was loaded */
         assert(end_index >= first_index);
-        int nb = snprintf(filename, sizeof(filename), UV__CLOSED_TEMPLATE, first_index, end_index);
+        int nb = snprintf(filename, sizeof(filename), UV__CLOSED_TEMPLATE,
+                          first_index, end_index);
         if ((nb < 0) || ((size_t)nb >= sizeof(filename))) {
             tracef("snprintf failed: %d", nb);
             rv = RAFT_IOERR;
@@ -615,7 +617,8 @@ done:
         info->first_index = first_index;
         info->end_index = end_index;
         memset(info->filename, '\0', sizeof(info->filename));
-        _Static_assert(sizeof(info->filename) >= sizeof(filename), "Destination buffer too small");
+        _Static_assert(sizeof(info->filename) >= sizeof(filename),
+                       "Destination buffer too small");
         /* info->filename is zeroed out, info->filename is at least as large as
          * filename and we checked that nb < sizeof(filename) -> we won't
          * overflow and the result will be zero terminated. */
@@ -696,7 +699,7 @@ void uvSegmentBufferClose(struct uvSegmentBuffer *b)
 int uvSegmentBufferFormat(struct uvSegmentBuffer *b)
 {
     int rv;
-    void *cursor;
+    uint8_t *cursor;
     size_t n;
     assert(b->n == 0);
     n = sizeof(uint64_t);
@@ -705,7 +708,7 @@ int uvSegmentBufferFormat(struct uvSegmentBuffer *b)
         return rv;
     }
     b->n = n;
-    cursor = b->arena.base;
+    cursor = (uint8_t *)b->arena.base;
     bytePut64(&cursor, UV__DISK_FORMAT);
     return 0;
 }
@@ -714,13 +717,13 @@ int uvSegmentBufferAppend(struct uvSegmentBuffer *b,
                           const struct raft_entry entries[],
                           unsigned n_entries)
 {
-    size_t size;   /* Total size of the batch */
-    uint32_t crc1; /* Header checksum */
-    uint32_t crc2; /* Data checksum */
-    void *crc1_p;  /* Pointer to header checksum slot */
-    void *crc2_p;  /* Pointer to data checksum slot */
-    void *header;  /* Pointer to the header section */
-    void *cursor;
+    size_t size;     /* Total size of the batch */
+    uint32_t crc1;   /* Header checksum */
+    uint32_t crc2;   /* Data checksum */
+    uint8_t *crc1_p; /* Pointer to header checksum slot */
+    uint8_t *crc2_p; /* Pointer to data checksum slot */
+    void *header;    /* Pointer to the header section */
+    uint8_t *cursor;
     unsigned i;
     int rv;
 
@@ -734,7 +737,7 @@ int uvSegmentBufferAppend(struct uvSegmentBuffer *b,
     if (rv != 0) {
         return rv;
     }
-    cursor = b->arena.base + b->n;
+    cursor = (uint8_t *)b->arena.base + b->n;
 
     /* Placeholder of the checksums */
     crc1_p = cursor;
@@ -752,8 +755,6 @@ int uvSegmentBufferAppend(struct uvSegmentBuffer *b,
     crc2 = 0;
     for (i = 0; i < n_entries; i++) {
         const struct raft_entry *entry = &entries[i];
-        /* TODO: enforce the requirement of 8-byte alignment also in the
-         * higher-level APIs. */
         assert(entry->buf.len % sizeof(uint64_t) == 0);
         memcpy(cursor, entry->buf.base, entry->buf.len);
         crc2 = byteCrc32(cursor, entry->buf.len, crc2);
@@ -807,11 +808,11 @@ void uvSegmentBufferReset(struct uvSegmentBuffer *b, unsigned retain)
  * Upon a restart, raft will not detect the segment anymore and will try
  * to start without it.
  * */
-#define CORRUPT_FILE_FMT "corrupt-%"PRId64"-%s"
+#define CORRUPT_FILE_FMT "corrupt-%" PRId64 "-%s"
 static void uvMoveCorruptSegment(struct uv *uv, struct uvSegmentInfo *info)
 {
     char errmsg[RAFT_ERRMSG_BUF_SIZE] = {0};
-    char new_filename[UV__FILENAME_LEN+1] = {0};
+    char new_filename[UV__FILENAME_LEN + 1] = {0};
     size_t sz = sizeof(new_filename);
     int rv;
 
@@ -835,12 +836,13 @@ static void uvMoveCorruptSegment(struct uv *uv, struct uvSegmentInfo *info)
 /*
  * On startup, raft will try to recover when a corrupt segment is detected.
  *
- * When a corrupt open segment is encountered, it, and all subsequent open segments,
- * are renamed. Not renaming newer, possible non-corrupt, open segments could lead
- * to loading inconsistent data.
+ * When a corrupt open segment is encountered, it, and all subsequent open
+ * segments, are renamed. Not renaming newer, possible non-corrupt, open
+ * segments could lead to loading inconsistent data.
  *
  * When a corrupt closed segment is encountered, it will be renamed when
- * it is the last closed segment, in that case all open-segments are renamed too.
+ * it is the last closed segment, in that case all open-segments are renamed
+ * too.
  */
 static void uvRecoverFromCorruptSegment(struct uv *uv,
                                         size_t i_corrupt,
@@ -968,7 +970,8 @@ err:
 static int uvWriteClosedSegment(struct uv *uv,
                                 raft_index first_index,
                                 raft_index last_index,
-                                const struct raft_buffer *conf)
+                                const struct raft_buffer *conf,
+                                raft_term conf_term)
 {
     char filename[UV__FILENAME_LEN];
     struct uvSegmentBuffer buf = {0};
@@ -999,7 +1002,7 @@ static int uvWriteClosedSegment(struct uv *uv,
         return rv;
     }
 
-    entry.term = 1;
+    entry.term = conf_term;
     entry.type = RAFT_CHANGE;
     entry.buf = *conf;
 
@@ -1024,13 +1027,14 @@ static int uvWriteClosedSegment(struct uv *uv,
 int uvSegmentCreateFirstClosed(struct uv *uv,
                                const struct raft_configuration *configuration)
 {
-    return uvSegmentCreateClosedWithConfiguration(uv, 1, configuration);
+    return uvSegmentCreateClosedWithConfiguration(uv, 1, configuration, 1);
 }
 
 int uvSegmentCreateClosedWithConfiguration(
     struct uv *uv,
     raft_index index,
-    const struct raft_configuration *configuration)
+    const struct raft_configuration *configuration,
+    raft_term conf_term)
 {
     struct raft_buffer buf;
     char filename[UV__FILENAME_LEN];
@@ -1046,7 +1050,7 @@ int uvSegmentCreateClosedWithConfiguration(
     }
 
     /* Write the file */
-    rv = uvWriteClosedSegment(uv, index, index, &buf);
+    rv = uvWriteClosedSegment(uv, index, index, &buf, conf_term);
     if (rv != 0) {
         goto err_after_configuration_encode;
     }
