@@ -1,5 +1,5 @@
-#include "uv_ip.h"
 #include "uv_tcp.h"
+#include "uv_ip.h"
 
 #include <string.h>
 
@@ -15,16 +15,10 @@ static int uvTcpInit(struct raft_uv_transport *transport,
                      const char *address)
 {
     struct UvTcp *t = transport->impl;
-    int rv;
     assert(id > 0);
     assert(address != NULL);
     t->id = id;
     t->address = address;
-    rv = uv_tcp_init(t->loop, &t->listener);
-    if (rv != 0) {
-        return rv;
-    }
-    t->listener.data = t;
     return 0;
 }
 
@@ -38,6 +32,7 @@ static void uvTcpClose(struct raft_uv_transport *transport,
     t->close_cb = cb;
     UvTcpListenClose(t);
     UvTcpConnectClose(t);
+    UvTcpMaybeFireCloseCb(t);
 }
 
 void UvTcpMaybeFireCloseCb(struct UvTcp *t)
@@ -48,11 +43,11 @@ void UvTcpMaybeFireCloseCb(struct UvTcp *t)
 
     assert(QUEUE_IS_EMPTY(&t->accepting));
     assert(QUEUE_IS_EMPTY(&t->connecting));
-
-    if (t->listener.data != NULL) {
+    if (!QUEUE_IS_EMPTY(&t->aborting)) {
         return;
     }
-    if (!QUEUE_IS_EMPTY(&t->aborting)) {
+
+    if (t->listeners != NULL) {
         return;
     }
 
@@ -66,8 +61,15 @@ int raft_uv_tcp_init(struct raft_uv_transport *transport,
 {
     struct UvTcp *t;
     void *data = transport->data;
+    int version = transport->version;
+    if (version != 1) {
+        ErrMsgPrintf(transport->errmsg, "Invalid version: %d", version);
+        return RAFT_INVALID;
+    }
+
     memset(transport, 0, sizeof *transport);
     transport->data = data;
+    transport->version = version;
     t = raft_malloc(sizeof *t);
     if (t == NULL) {
         ErrMsgOom(transport->errmsg);
@@ -78,7 +80,8 @@ int raft_uv_tcp_init(struct raft_uv_transport *transport,
     t->id = 0;
     t->address = NULL;
     t->bind_address = NULL;
-    t->listener.data = NULL;
+    t->listeners = NULL;
+    t->n_listeners = 0;
     t->accept_cb = NULL;
     QUEUE_INIT(&t->accepting);
     QUEUE_INIT(&t->connecting);
@@ -106,10 +109,12 @@ int raft_uv_tcp_set_bind_address(struct raft_uv_transport *transport,
                                  const char *address)
 {
     struct UvTcp *t = transport->impl;
-    struct sockaddr_in addr;
+    char hostname[NI_MAXHOST];
+    char service[NI_MAXSERV];
     int rv;
 
-    rv = uvIpParse(address, &addr);
+    rv = uvIpAddrSplit(address, hostname, sizeof(hostname), service,
+                       sizeof(service));
     if (rv != 0) {
         return RAFT_INVALID;
     }
