@@ -10,13 +10,22 @@
 /* The happy path of a connection request is:
  *
  * - Create a TCP handle and submit a TCP connect request.
+ * - Initiate an asynchronous dns resolve request
+ * - Once the name lookup was successfull connect to the first resolved IP
  * - Once connected over TCP, submit a write request for the handshake.
  * - Once the write completes, fire the connection request callback.
  *
+ * Alternative happy path of a connection request, if hostname resolves to
+ * multiple IPs and first/second/... IP is reachable:
+ * - close the tcp handle and initiate a new connect with next IP in cb
+ *
  * Possible failure modes are:
  *
- * - The transport get closed, close the TCP handle and and fire the request
- *   callback with RAFT_CANCELED.
+ * - The name resolve for the hostname is not sucessfull, close the TCP handle
+ *   and fire the request callback.
+ *
+ * - The raft_uv_transport object gets closed, close the TCP handle and and fire
+ *   the request callback with RAFT_CANCELED.
  *
  * - Either the TCP connect or the write request fails: close the TCP handle and
  *   fire the request callback with RAFT_NOCONNECTION.
@@ -25,34 +34,38 @@
 /* Hold state for a single connection request. */
 struct uvTcpConnect
 {
-    struct UvTcp *t;             /* Transport implementation */
-    struct raft_uv_connect *req; /* User request */
-    uv_buf_t handshake;          /* Handshake data */
-    struct uv_tcp_s *tcp;        /* TCP connection socket handle */
-    struct uv_connect_s connect; /* TCP connection request */
-    struct uv_write_s write;     /* TCP handshake request */
-    int status;                  /* Returned to the request callback */
-    queue queue;                 /* Pending connect queue */
+    struct UvTcp *t;                     /* Transport implementation */
+    struct raft_uv_connect *req;         /* User request */
+    uv_buf_t handshake;                  /* Handshake data */
+    struct uv_tcp_s *tcp;                /* TCP connection socket handle */
+    struct uv_getaddrinfo_s getaddrinfo; /* DNS resolve request */
+    const struct addrinfo *ai_current; /* The current sockaddr to connect to */
+    struct uv_connect_s connect;       /* TCP connection request */
+    struct uv_write_s write;           /* TCP handshake request */
+    int status;                        /* Returned to the request callback */
+    bool resolving;                    /* Indicate name resolving in progress */
+    bool retry;  /* Indicate tcp connect failure handling */
+    queue queue; /* Pending connect queue */
 };
 
 /* Encode an handshake message into the given buffer. */
 static int uvTcpEncodeHandshake(raft_id id, const char *address, uv_buf_t *buf)
 {
-    void *cursor;
+    uint8_t *cursor;
     size_t address_len = bytePad64(strlen(address) + 1);
     buf->len = sizeof(uint64_t) + /* Protocol version. */
                sizeof(uint64_t) + /* Server ID. */
                sizeof(uint64_t) /* Size of the address buffer */;
     buf->len += address_len;
-    buf->base = HeapMalloc(buf->len);
+    buf->base = RaftHeapMalloc(buf->len);
     if (buf->base == NULL) {
         return RAFT_NOMEM;
     }
-    cursor = buf->base;
+    cursor = (uint8_t *)buf->base;
     bytePut64(&cursor, UV__TCP_HANDSHAKE_PROTOCOL);
     bytePut64(&cursor, id);
     bytePut64(&cursor, address_len);
-    strcpy(cursor, address);
+    strcpy((char *)cursor, address);
     return 0;
 }
 
@@ -64,7 +77,8 @@ static void uvTcpConnectFinish(struct uvTcpConnect *connect)
     struct raft_uv_connect *req = connect->req;
     int status = connect->status;
     QUEUE_REMOVE(&connect->queue);
-    HeapFree(connect->handshake.base);
+    RaftHeapFree(connect->handshake.base);
+    uv_freeaddrinfo(connect->getaddrinfo.addrinfo);
     raft_free(connect);
     req->cb(req, stream, status);
 }
@@ -77,7 +91,7 @@ static void uvTcpConnectUvCloseCb(struct uv_handle_s *handle)
     struct UvTcp *t = connect->t;
     assert(connect->status != 0);
     assert(handle == (struct uv_handle_s *)connect->tcp);
-    HeapFree(connect->tcp);
+    RaftHeapFree(connect->tcp);
     connect->tcp = NULL;
     uvTcpConnectFinish(connect);
     UvTcpMaybeFireCloseCb(t);
@@ -88,7 +102,12 @@ static void uvTcpConnectAbort(struct uvTcpConnect *connect)
 {
     QUEUE_REMOVE(&connect->queue);
     QUEUE_PUSH(&connect->t->aborting, &connect->queue);
-    uv_close((struct uv_handle_s *)connect->tcp, uvTcpConnectUvCloseCb);
+    uv_cancel((struct uv_req_s *)&connect->getaddrinfo);
+    /* If there is no getaddrinfo request in flight, close the TCP handle now
+     * (otherwise it will be closed after the getaddrinfo request completes). */
+    if (!connect->resolving && !connect->retry) {
+        uv_close((struct uv_handle_s *)connect->tcp, uvTcpConnectUvCloseCb);
+    }
 }
 
 /* The handshake TCP write completes. Fire the connect callback. */
@@ -112,6 +131,30 @@ static void uvTcpConnectUvWriteCb(struct uv_write_s *write, int status)
     uvTcpConnectFinish(connect);
 }
 
+/* Helper function to connect to the remote node */
+static void uvTcpAsyncConnect(struct uvTcpConnect *connect);
+
+/* The TCP connect failed, we closed the handle and want to try with next IP */
+static void uvTcpTryNextConnectCb(struct uv_handle_s *handle)
+{
+    struct uvTcpConnect *connect = handle->data;
+    struct UvTcp *t = connect->t;
+    int rv;
+
+    connect->retry = false;
+
+    if (t->closing) {
+        connect->status = RAFT_CANCELED;
+        /* We are already in close cb for the tcp handle, simply invoke final cb
+         */
+        uvTcpConnectUvCloseCb(handle);
+        return;
+    }
+    rv = uv_tcp_init(t->loop, connect->tcp);
+    assert(rv == 0);
+    uvTcpAsyncConnect(connect);
+}
+
 /* The TCP connection is established. Write the handshake data. */
 static void uvTcpConnectUvConnectCb(struct uv_connect_s *req, int status)
 {
@@ -126,6 +169,15 @@ static void uvTcpConnectUvConnectCb(struct uv_connect_s *req, int status)
 
     if (status != 0) {
         assert(status != UV_ECANCELED); /* t->closing would have been true */
+        connect->ai_current = connect->ai_current->ai_next;
+        if (connect->ai_current) {
+            /* For the next connect attempt we need to close the tcp handle. */
+            /* To avoid interference with aborting we set a flag to indicate the
+             * connect attempt */
+            connect->retry = true;
+            uv_close((struct uv_handle_s *)connect->tcp, uvTcpTryNextConnectCb);
+            return;
+        }
         connect->status = RAFT_NOCONNECTION;
         ErrMsgPrintf(t->transport->errmsg, "uv_tcp_connect(): %s",
                      uv_strerror(status));
@@ -146,15 +198,71 @@ err:
     uvTcpConnectAbort(connect);
 }
 
+/* Helper function to connect to the remote node */
+static void uvTcpAsyncConnect(struct uvTcpConnect *connect)
+{
+    int rv;
+    rv = uv_tcp_connect(&connect->connect, connect->tcp,
+                        connect->ai_current->ai_addr, uvTcpConnectUvConnectCb);
+    if (rv != 0) {
+        /* UNTESTED: since parsing succeed, this should fail only because of
+         * lack of system resources */
+        ErrMsgPrintf(connect->t->transport->errmsg, "uv_tcp_connect(): %s",
+                     uv_strerror(rv));
+        connect->status = RAFT_NOCONNECTION;
+        uvTcpConnectAbort(connect);
+    }
+}
+
+/* The hostname resolve is finished */
+static void uvTcpConnectGetAddrInfoCb(uv_getaddrinfo_t *req,
+                                      int status,
+                                      struct addrinfo *res)
+{
+    struct uvTcpConnect *connect = req->data;
+    struct UvTcp *t = connect->t;
+
+    connect->resolving =
+        false; /* Indicate we are in the name resolving phase */
+
+    if (t->closing) {
+        connect->status = RAFT_CANCELED;
+
+        /* We need to close the tcp handle to abort connection attempt */
+        uv_close((struct uv_handle_s *)connect->tcp, uvTcpConnectUvCloseCb);
+        return;
+    }
+
+    if (status < 0) {
+        ErrMsgPrintf(t->transport->errmsg, "uv_getaddrinfo(): %s",
+                     uv_err_name(status));
+        connect->status = RAFT_NOCONNECTION;
+        uvTcpConnectAbort(connect);
+        return;
+    }
+    connect->ai_current = res;
+    uvTcpAsyncConnect(connect);
+}
+
 /* Create a new TCP handle and submit a connection request to the event loop. */
 static int uvTcpConnectStart(struct uvTcpConnect *r, const char *address)
 {
     struct UvTcp *t = r->t;
-    struct sockaddr_in addr;
+    static struct addrinfo hints = {.ai_family = AF_INET,
+                                    .ai_socktype = SOCK_STREAM,
+                                    .ai_protocol = 0,
+                                    .ai_flags = 0};
+    char hostname[NI_MAXHOST];
+    char service[NI_MAXSERV];
     int rv;
 
-    rv = uvIpParse(address, &addr);
-    if (rv != 0) {
+    rv = uvIpAddrSplit(address, hostname, sizeof(hostname), service,
+                       sizeof(service));
+    if (rv) {
+        ErrMsgPrintf(t->transport->errmsg,
+                     "uv_tcp_connect(): Cannot split %s into host and service",
+                     address);
+        rv = RAFT_NOCONNECTION;
         goto err;
     }
 
@@ -162,11 +270,11 @@ static int uvTcpConnectStart(struct uvTcpConnect *r, const char *address)
     rv = uvTcpEncodeHandshake(t->id, t->address, &r->handshake);
     if (rv != 0) {
         assert(rv == RAFT_NOMEM);
-        ErrMsgOom(r->t->transport->errmsg);
+        ErrMsgOom(t->transport->errmsg);
         goto err;
     }
 
-    r->tcp = HeapMalloc(sizeof *r->tcp);
+    r->tcp = RaftHeapMalloc(sizeof *r->tcp);
     if (r->tcp == NULL) {
         ErrMsgOom(t->transport->errmsg);
         rv = RAFT_NOMEM;
@@ -177,23 +285,23 @@ static int uvTcpConnectStart(struct uvTcpConnect *r, const char *address)
     assert(rv == 0);
     r->tcp->data = r;
 
-    rv = uv_tcp_connect(&r->connect, r->tcp, (struct sockaddr *)&addr,
-                        uvTcpConnectUvConnectCb);
-    if (rv != 0) {
-        /* UNTESTED: since parsing succeed, this should fail only because of
-         * lack of system resources */
-        ErrMsgPrintf(t->transport->errmsg, "uv_tcp_connect(): %s",
+    rv = uv_getaddrinfo(r->t->loop, &r->getaddrinfo, &uvTcpConnectGetAddrInfoCb,
+                        hostname, service, &hints);
+    if (rv) {
+        ErrMsgPrintf(t->transport->errmsg,
+                     "uv_tcp_connect(): Cannot initiate getaddrinfo %s",
                      uv_strerror(rv));
         rv = RAFT_NOCONNECTION;
         goto err_after_tcp_init;
     }
+    r->resolving = true; /* Indicate we are in the name resolving phase */
 
     return 0;
 
 err_after_tcp_init:
-    uv_close((uv_handle_t *)r->tcp, (uv_close_cb)HeapFree);
+    uv_close((uv_handle_t *)r->tcp, (uv_close_cb)RaftHeapFree);
 err_after_encode_handshake:
-    HeapFree(r->handshake.base);
+    RaftHeapFree(r->handshake.base);
 err:
     return rv;
 }
@@ -211,7 +319,7 @@ int UvTcpConnect(struct raft_uv_transport *transport,
     assert(!t->closing);
 
     /* Create and initialize a new TCP connection request object */
-    r = HeapMalloc(sizeof *r);
+    r = RaftHeapMalloc(sizeof *r);
     if (r == NULL) {
         rv = RAFT_NOMEM;
         ErrMsgOom(transport->errmsg);
@@ -221,8 +329,10 @@ int UvTcpConnect(struct raft_uv_transport *transport,
     r->req = req;
     r->status = 0;
     r->write.data = r;
+    r->getaddrinfo.data = r;
+    r->resolving = false;
+    r->retry = false;
     r->connect.data = r;
-
     req->cb = cb;
 
     /* Keep track of the pending request */
@@ -238,7 +348,7 @@ int UvTcpConnect(struct raft_uv_transport *transport,
 
 err_after_alloc:
     QUEUE_REMOVE(&r->queue);
-    HeapFree(r);
+    RaftHeapFree(r);
 err:
     return rv;
 }
